#  **************************************************************************  #
"""
#   BornAgain: simulate and fit reflection and scattering
#
#   @file      Wrap/Python/ba_fitmonitor.py
#   @brief     Plotter classes for monitoring fit progress.
#
#   @homepage  http://apps.jcns.fz-juelich.de/BornAgain
#   @license   GNU General Public License v3 or higher (see COPYING)
#   @copyright Forschungszentrum Juelich GmbH 2019
#   @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
"""
#  **************************************************************************  #

import bornagain as ba
import ba_plot as bp
try:  # workaround for build servers
    import numpy as np
    from matplotlib import pyplot as plt
    from matplotlib import gridspec, colors
except Exception as e:
    print("In ba_fitmonitor.py: {:s}".format(str(e)))

label_fontsize = 16


class Plotter:
    """
    Draws fit progress. Base class for simulation-specific classes (PlotterGISAS etc).
    """
    def __init__(self,
                 zmin=None,
                 zmax=None,
                 xlabel=None,
                 ylabel=None,
                 units=ba.Axes.DEFAULT,
                 aspect=None):

        self._fig = plt.figure(figsize=(10.25, 7.69))
        self._fig.canvas.draw()
        self._zmin = zmin
        self._zmax = zmax
        self._xlabel = xlabel
        self._ylabel = ylabel
        self._units = units
        self._aspect = aspect

    def __call__(self, fit_objective):
        self.plot(fit_objective)

    def reset(self):
        self._fig.clf()

    def plot(self):
        self._fig.tight_layout()
        plt.pause(0.03)


class PlotterGISAS(Plotter):
    """
    Draws fit progress, for GISAS simulation.
    """
    def __init__(self,
                 zmin=None,
                 zmax=None,
                 xlabel=None,
                 ylabel=None,
                 units=ba.Axes.DEFAULT,
                 aspect=None):
        Plotter.__init__(self, zmin, zmax, xlabel, ylabel, units, aspect)

    @staticmethod
    def make_subplot(nplot):
        plt.subplot(2, 2, nplot)
        plt.subplots_adjust(wspace=0.2, hspace=0.2)

    def plot(self, fit_objective):
        Plotter.reset(self)

        real_data = fit_objective.experimentalData()
        sim_data = fit_objective.simulationResult()
        diff = fit_objective.absoluteDifference()

        self.make_subplot(1)

        # same limits for both plots
        arr = real_data.array()
        zmax = np.amax(arr) if self._zmax is None else self._zmax
        zmin = zmax*1e-6 if self._zmin is None else self._zmin

        bp.plot_colormap(real_data,
                         title="Experimental data",
                         intensity_min=zmin,
                         intensity_max=zmax,
                         units=self._units,
                         xlabel=self._xlabel,
                         ylabel=self._ylabel,
                         zlabel='',
                         aspect=self._aspect)

        self.make_subplot(2)
        bp.plot_colormap(sim_data,
                         title="Simulated data",
                         intensity_min=zmin,
                         intensity_max=zmax,
                         units=self._units,
                         xlabel=self._xlabel,
                         ylabel=self._ylabel,
                         zlabel='',
                         aspect=self._aspect)

        self.make_subplot(3)
        bp.plot_colormap(diff,
                         title="Difference",
                         intensity_min=zmin,
                         intensity_max=zmax,
                         units=self._units,
                         xlabel=self._xlabel,
                         ylabel=self._ylabel,
                         zlabel='',
                         aspect=self._aspect)

        self.make_subplot(4)
        plt.title('Parameters')
        plt.axis('off')

        iteration_info = fit_objective.iterationInfo()

        plt.text(
            0.01, 0.85, "Iterations  " +
            '{:d}'.format(iteration_info.iterationCount()))
        plt.text(0.01, 0.75,
                 "Chi2       " + '{:8.4f}'.format(iteration_info.chi2()))
        index = 0
        params = iteration_info.parameterMap()
        for key in params:
            plt.text(0.01, 0.55 - index*0.1,
                     '{:30.30s}: {:6.3f}'.format(key, params[key]))
            index = index + 1

        Plotter.plot(self)


class PlotterSpecular(Plotter):
    """
    Draws fit progress, for specular simulation.
    """
    def __init__(self, units=ba.Axes.DEFAULT):
        Plotter.__init__(self)
        self.gs = gridspec.GridSpec(1, 2, width_ratios=[2.5, 1], wspace=0)
        self.units = units

    @staticmethod
    def as_si(val, ndp):
        """
        Fancy print of scientific-formatted values
        :param val: numeric value
        :param ndp: number of decimal digits to print
        :return: a string corresponding to the _val_
        """
        s = '{x:0.{ndp:d}e}'.format(x=val, ndp=ndp)
        m, e = s.split('e')
        return r'{m:s}\times 10^{{{e:d}}}'.format(m=m, e=int(e))

    @staticmethod
    def trunc_str(token, length):
        """
        Truncates token if it is longer than length.

        Example:
            trunc_str("123456789", 8) returns "123456.."

            trunc_str("123456789", 9) returns "123456789"

        :param token: input string
        :param length: max non-truncated length
        :return:
        """
        return (token[:length - 2] +
                '..') if len(token) > length else token

    def plot_table(self, fit_objective):

        iteration_info = fit_objective.iterationInfo()

        trunc_length = 9  # max string field width in the table
        n_digits = 1  # number of decimal digits to print

        n_iterations = iteration_info.iterationCount(
        )  # current number of iterations passed
        rel_dif = fit_objective.relativeDifference().array().max(
        )  # maximum relative difference
        fitted_parameters = iteration_info.parameterMap()

        # creating table content
        labels = ("Parameter", "Value")
        table_data = [["Iteration", '${:d}$'.format(n_iterations)],
                      [
                          "$d_{r, max}$",
                          '${:s}$'.format(self.as_si(rel_dif, n_digits))
                      ]]
        for key, value in fitted_parameters.iteritems():
            table_data.append([
                '{:s}'.format(self.trunc_str(key, trunc_length)),
                '${:s}$'.format(self.as_si(value, n_digits))
            ])

        # creating table
        axs = plt.subplot(self.gs[1])
        axs.axis('tight')
        axs.axis('off')
        table = plt.table(cellText=table_data,
                          colLabels=labels,
                          cellLoc='center',
                          loc='bottom left',
                          bbox=[0, 0, 1, 1.0])

    def plot_graph(self, fit_objective):
        # retrieving data from fit suite
        real_data = fit_objective.experimentalData()
        sim_data = fit_objective.simulationResult()
        unc_data = fit_objective.uncertaintyData()

        # data values
        sim_values = sim_data.array(self.units)
        real_values = real_data.array(self.units)
        unc_values = None if unc_data is None else unc_data.array(
            self.units)

        # default font properties dictionary to use
        font = {
            'family': 'serif',
            'weight': 'normal',
            'size': label_fontsize
        }

        plt.subplot(self.gs[0])
        plt.semilogy(sim_data.axis(), sim_values, 'b', real_data.axis(),
                     real_values, 'k--')
        if unc_values is not None:
            plt.semilogy(real_data.axis(),
                         real_values - unc_values,
                         'xkcd:grey',
                         alpha=0.6)
            plt.semilogy(real_data.axis(),
                         real_values + unc_values,
                         'xkcd:grey',
                         alpha=0.6)
        plt.ylim((0.5*np.min(real_values), 5*np.max(real_values)))

        xlabel = bp.get_axes_labels(real_data, self.units)[0]
        legend = ['BornAgain', 'Data']
        if unc_values is not None:
            legend = ['BornAgain', 'Data', r'Data $\pm \sigma$']
        plt.legend(legend, loc='upper right', prop=font)
        plt.xlabel(xlabel, fontdict=font)
        plt.ylabel("Intensity", fontdict=font)
        plt.title("Specular data fitting", fontdict=font)

    def plot(self, fit_objective):
        Plotter.reset(self)

        self.plot_graph(fit_objective)
        self.plot_table(fit_objective)

        Plotter.plot(self)



#!/usr/bin/env python3
import bornagain as ba
from bornagain import deg, nm


def get_sample():
    """
    Returns a sample with cylinders on a substrate,
    forming a 2D lattice with different disorder rotated lattice
    """
    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_substrate = ba.HomogeneousMaterial("Substrate", 6e-6, 2e-8)
    m_particle = ba.HomogeneousMaterial("Particle", 6e-4, 2e-8)

    vacuum_layer = ba.Layer(m_vacuum)
    substrate_layer = ba.Layer(m_substrate)

    p_interference_function = \
        ba.InterferenceFunction2DLattice(ba.SquareLattice2D(25*nm, 0))
    pdf = ba.FTDecayFunction2DCauchy(48*nm, 16*nm, 0)
    p_interference_function.setDecayFunction(pdf)

    particle_layout = ba.ParticleLayout()
    ff = ba.FormFactorCylinder(3*nm, 3*nm)
    position = ba.kvector_t(0, 0, 0)
    cylinder = ba.Particle(m_particle, ff.clone())
    cylinder.setPosition(position)
    particle_layout.addParticle(cylinder)
    particle_layout.setInterferenceFunction(p_interference_function)

    vacuum_layer.addLayout(particle_layout)

    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    multi_layer.addLayer(substrate_layer)
    return multi_layer


def get_simulation(sample):
    beam = ba.Beam(1, 0.1*nm, ba.Direction(0.2*deg, 0))
    detector = ba.SphericalDetector(100, 2*deg, 1*deg, 1*deg)
    simulation = ba.GISASSimulation(beam, sample, detector)
    distr_1 = ba.DistributionGate(0, 240*deg)
    simulation.addParameterDistribution("*/SquareLattice2D/Xi", distr_1, 3,
                                        0)
    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)



# Functional test: tests of IO operations with the IntensityData object

import math, numpy, os, sys, time, unittest
import bornagain as ba


def fill_data(data):
    """
    Fills intensity data with some numbers
    """
    for i in range(0, data.getAllocatedSize()):
        data[i] = i


def is_the_same_data(data1, data2):
    """
    Checks if two data are identical
    """
    if data1.getAllocatedSize() != data2.getAllocatedSize():
        return False
    if data1.rank() != data2.rank():
        return False
    for i in range(0, data1.rank()):
        if data1.axis(i) != data2.axis(i):
            return False
    for i in range(0, data1.getAllocatedSize()):
        if data1[i] != data2[i]:
            return False
    return True


class OutputDataIOTiffTest(unittest.TestCase):
    """
    Test serialization of ba.IntensityData into TIFF format
    """
    def test_SaveToTiff(self):
        data = ba.IntensityData()
        data.addAxis(ba.FixedBinAxis("x", 10, 0, 10))
        data.addAxis(ba.FixedBinAxis("y", 5, 0, 5))
        fill_data(data)

        ba.IntensityDataIOFactory.writeOutputData(data, "tmp.tif")
        newdata = ba.IntensityDataIOFactory.readOutputData("tmp.tif")
        self.assertTrue(is_the_same_data(data, newdata))

        ba.IntensityDataIOFactory.writeOutputData(data, "tmp.tif.gz")
        newdata = ba.IntensityDataIOFactory.readOutputData("tmp.tif.gz")
        self.assertTrue(is_the_same_data(data, newdata))

        ba.IntensityDataIOFactory.writeOutputData(data, "tmp.tif.bz2")
        newdata = ba.IntensityDataIOFactory.readOutputData("tmp.tif.bz2")
        self.assertTrue(is_the_same_data(data, newdata))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
Basic example for regular small-angle scattering (SAS).
Sample is a dilute assembly of ordered dodecahedra.
"""

import bornagain as ba
from bornagain import deg, nm


def get_sample():
    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_solution = ba.HomogeneousMaterial("Solution", 6e-6, 2e-8)
    m_particle = ba.HomogeneousMaterial("Particle", 6e-4, 2e-8)

    # Finite sample layer, contains particles in solution:
    ff = ba.FormFactorDodecahedron(12*nm)
    particle = ba.Particle(m_particle, ff)
    layout = ba.ParticleLayout()
    layout.addParticle(particle)
    solution_layer = ba.Layer(m_solution, 1000*nm)
    # TODO: make intensity proportional to thickness,
    #       https://github.com/scgmlz/BornAgain/issues/1222
    solution_layer.addLayout(layout)

    # Flat sample layer sandwiched between semi-infinite vacuum layers:
    sample = ba.MultiLayer()
    sample.addLayer(ba.Layer(m_vacuum))
    sample.addLayer(solution_layer)
    sample.addLayer(ba.Layer(m_vacuum))
    return sample


def get_simulation(sample):
    # Beam from above (perpendicular to sample):
    beam = ba.Beam(1, 0.4*nm, ba.Direction(90*deg, 0))

    # Detector opposite to source:
    detPos = 2000  # distance from sample center to detector in mm
    detWid = 500  # detector width in mm
    detPix = 200  # number of pixels per direction
    det = ba.RectangularDetector(detPix, detWid, detPix, detWid)
    det.setPerpendicularToDirectBeam(detPos, detWid/2, detWid/2)

    return ba.GISASSimulation(beam, sample, det)


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)



#!/usr/bin/env python3
"""
Fitting example: running one fit after another using different minimizers.

During the first fit Genetic minimizer will be used to scan large parameter
space and roughly find local minimum. The second Minuit2 minimizer will continue
after that to find precise minimum location.
"""

import numpy as np
from matplotlib import pyplot as plt
import bornagain as ba
import ba_fitmonitor
from bornagain import deg, angstrom, nm


def get_sample(params):
    """
    Returns a sample with uncorrelated cylinders and pyramids on a substrate.
    """
    radius = params["radius"]
    height = params["height"]

    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_substrate = ba.HomogeneousMaterial("Substrate", 6e-6, 2e-8)
    m_particle = ba.HomogeneousMaterial("Particle", 6e-4, 2e-8)

    cylinder_ff = ba.FormFactorCylinder(radius, height)
    cylinder = ba.Particle(m_particle, cylinder_ff)

    layout = ba.ParticleLayout()
    layout.addParticle(cylinder)

    vacuum_layer = ba.Layer(m_vacuum)
    vacuum_layer.addLayout(layout)

    substrate_layer = ba.Layer(m_substrate, 0)
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    multi_layer.addLayer(substrate_layer)
    return multi_layer


def get_simulation(params):
    """
    Returns a GISAXS simulation with beam and detector defined.
    """
    simulation = ba.GISASSimulation()
    simulation.setDetectorParameters(100, 0, 2*deg, 100, 0, 2*deg)
    simulation.setBeamParameters(1*angstrom, 0.2*deg, 0)
    simulation.beam().setIntensity(1e+08)
    simulation.setSample(get_sample(params))
    return simulation


def create_real_data():
    """
    Generating "real" data by adding noise to the simulated data.
    """
    params = {'radius': 5*nm, 'height': 5*nm}

    simulation = get_simulation(params)
    simulation.runSimulation()

    # retrieving simulated data in the form of numpy array
    real_data = simulation.result().array()

    # spoiling simulated data with the noise to produce "real" data
    noise_factor = 0.1
    noisy = np.random.normal(real_data, noise_factor*np.sqrt(real_data))
    noisy[noisy < 0.1] = 0.1
    return noisy


def run_fitting():
    """
    main function to run fitting
    """
    real_data = create_real_data()

    fit_objective = ba.FitObjective()
    fit_objective.addSimulationAndData(get_simulation, real_data, 1)
    fit_objective.initPrint(10)
    observer = ba_fitmonitor.PlotterGISAS()
    fit_objective.initPlot(10, observer)
    """
    Setting fitting parameters with starting values.
    Here we select starting values being quite far from true values
    to puzzle our minimizer's as much as possible.
    """
    params = ba.Parameters()
    params.add("height", 1.*nm, min=0.01, max=30, step=0.05*nm)
    params.add("radius", 20.*nm, min=0.01, max=30, step=0.05*nm)
    """
    Now we run first minimization round using the Genetic minimizer.
    The Genetic minimizer is able to explore large parameter space
    without being trapped by some local minimum.
    """
    minimizer = ba.Minimizer()
    minimizer.setMinimizer("Genetic", "", "MaxIterations=2;RandomSeed=1")
    result = minimizer.minimize(fit_objective.evaluate, params)
    fit_objective.finalize(result)

    best_params_so_far = result.parameters()
    """
    Second fit will use another minimizer.
    It starts from best parameters found in previous minimization
    and then continues until fit converges.
    """
    minimizer.setMinimizer("Minuit2", "Migrad")
    result = minimizer.minimize(fit_objective.evaluate, best_params_so_far)

    fit_objective.finalize(result)


if __name__ == '__main__':
    run_fitting()
    plt.show()



#!/usr/bin/env python3
"""
An example of computing splin-flip reflectivity from
a magnetized sample.
"""
import numpy
import matplotlib.pyplot as plt

import bornagain as ba
from bornagain import angstrom, deg, nm, nm2, kvector_t
from math import sqrt

def get_sample():
    """
    Defines sample and returns it
    """

    # Define materials
    material_Ambient = ba.MaterialBySLD("Ambient", 0, 0)
    h = 1e8
    magnetic_field = kvector_t(1/2*h, sqrt(3)/2*h, 0)
    material_Layer = ba.MaterialBySLD("Layer", 0.0001, 1e-08,
                                      magnetic_field)
    material_Substrate = ba.MaterialBySLD("Substrate", 7e-05, 2e-06)

    # Define layers
    layer_1 = ba.Layer(material_Ambient)
    layer_2 = ba.Layer(material_Layer, 10*nm)
    layer_3 = ba.Layer(material_Substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)
    sample.addLayer(layer_3)

    return sample


def get_simulation(sample, scan_size=500):
    """
    Defines and returns a specular simulation.
    """
    simulation = ba.SpecularSimulation()
    scan = ba.AngularSpecScan(1.54*angstrom, scan_size, 0, 5*deg)
    simulation.setScan(scan)
    simulation.setSample(sample)
    return simulation


def run_simulation(polarization=ba.kvector_t(0, 1, 0), analyzer=None):
    """
    Runs simulation and returns its result.
    """
    sample = get_sample()
    simulation = get_simulation(sample)

    # adding polarization and analyzer operator
    simulation.beam().setPolarization(polarization)
    if analyzer:
        simulation.detector().setAnalyzerProperties(analyzer, 1, 0.5)

    simulation.runSimulation()
    result = simulation.result()
    return result.axis(), result.array()


def plot(axis, data, labels):

    plt.figure()
    for d, l in zip(data, labels):
        plt.semilogy(axis, d, label=l, linewidth=1)

    plt.legend(loc='upper right')
    plt.gca().yaxis.set_ticks_position('both')
    plt.gca().xaxis.set_ticks_position('both')

    plt.xlabel(r"$\alpha_i$ [deg]")
    plt.ylabel("Reflectivity")

    plt.tight_layout()


if __name__ == '__main__':
    q, results_pp = run_simulation(ba.kvector_t(0, 1, 0),
                                   ba.kvector_t(0, 1, 0))
    q, results_mm = run_simulation(ba.kvector_t(0, -1, 0),
                                   ba.kvector_t(0, -1, 0))

    q, results_pm = run_simulation(ba.kvector_t(0, 1, 0),
                                   ba.kvector_t(0, -1, 0))
    q, results_mp = run_simulation(ba.kvector_t(0, -1, 0),
                                   ba.kvector_t(0, 1, 0))

    r_plus = results_pp + results_pm
    r_minus = results_mm + results_mp
    plot(q, [r_plus, r_minus], ["$+$", "$-$"])

    # same result, but need half the computational time
    q, results_p = run_simulation(ba.kvector_t(0, 1, 0))
    q, results_m = run_simulation(ba.kvector_t(0, -1, 0))

    plot(q, [results_p, results_m], ["$+$", "$-$"])

    plt.show()



#!/usr/bin/env python3
"""
This simulation example demonstrates how to replicate the
fitting example "Magnetically Dead Layers in Spinel Films"
given at the Nist website:
https://www.nist.gov/ncnr/magnetically-dead-layers-spinel-films

For simplicity, here we only reproduce the first part of that 
demonstration without the magnetically dead layer.
"""

# import boranagain
import bornagain as ba
from bornagain import deg, angstrom, nm

import numpy
import matplotlib.pyplot as plt

# import more libs needed for data processing
from re import match, DOTALL
from sys import argv
from io import BytesIO
from urllib.request import urlopen
from zipfile import ZipFile
from os.path import isfile

# q-range on which the simulation and fitting are to be performed
qmin = 0.05997
qmax = 1.96

# number of points on which the computed result is plotted
scan_size = 1500

# The SLD of the substrate is kept constant
sldMao = (5.377e-06, 0)

# constant to convert between magnetization and magnetic SLD
RhoMconst = 2.910429812376859e-12

####################################################################
#                  Create Sample and Simulation                    #
####################################################################


def get_sample(params):
    """
    construct the sample with the given parameters
    """
    magnetizationMagnitude = params["rhoM_Mafo"]*1e-6/RhoMconst
    angle = 0
    magnetizationVector = ba.kvector_t(
        magnetizationMagnitude*numpy.sin(angle*deg),
        magnetizationMagnitude*numpy.cos(angle*deg), 0)

    mat_vacuum = ba.MaterialBySLD("Vacuum", 0, 0)
    mat_layer = ba.MaterialBySLD("(Mg,Al,Fe)3O4", params["rho_Mafo"]*1e-6,
                                 0, magnetizationVector)
    mat_substrate = ba.MaterialBySLD("MgAl2O4", *sldMao)

    ambient_layer = ba.Layer(mat_vacuum)
    layer = ba.Layer(mat_layer, params["t_Mafo"]*angstrom)
    substrate_layer = ba.Layer(mat_substrate)

    r_Mafo = ba.LayerRoughness()
    r_Mafo.setSigma(params["r_Mafo"]*angstrom)

    r_substrate = ba.LayerRoughness()
    r_substrate.setSigma(params["r_Mao"]*angstrom)

    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(ambient_layer)
    multi_layer.addLayerWithTopRoughness(layer, r_Mafo)
    multi_layer.addLayerWithTopRoughness(substrate_layer, r_substrate)

    return multi_layer


def get_simulation(q_axis, parameters, polarization, analyzer):
    """
    Returns a simulation object.
    Polarization, analyzer and resolution are set
    from given parameters
    """
    simulation = ba.SpecularSimulation()
    q_axis = q_axis + parameters["q_offset"]
    scan = ba.QSpecScan(q_axis)

    dq = parameters["q_res"]*q_axis
    n_sig = 4.0
    n_samples = 25

    distr = ba.RangedDistributionGaussian(n_samples, n_sig)
    scan.setAbsoluteQResolution(distr, parameters["q_res"])

    simulation.beam().setPolarization(polarization)
    simulation.detector().setAnalyzerProperties(analyzer, 1, 0.5)

    simulation.setScan(scan)
    return simulation


def run_simulation(q_axis, fitParams, *, polarization, analyzer):
    """
    Run a simulation on the given q-axis, where the sample is 
    constructed with the given parameters.
    Vectors for polarization and analyzer need to be provided
    """
    parameters = dict(fitParams, **fixedParams)

    sample = get_sample(parameters)
    simulation = get_simulation(q_axis, parameters, polarization, analyzer)

    simulation.setSample(sample)
    simulation.runSimulation()
    return simulation


def qr(result):
    """
    Returns two arrays that hold the q-values as well as the
    reflectivity from a given simulation result
    """
    q = numpy.array(result.result().axis(ba.Axes.QSPACE))
    r = numpy.array(result.result().array(ba.Axes.QSPACE))

    return q, r


####################################################################
#                         Plot Handling                            #
####################################################################


def plot(qs, rs, exps, labels, filename):
    """
    Plot the simulated result together with the experimental data
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)

    for q, r, exp, l in zip(qs, rs, exps, labels):

        ax.errorbar(exp[0],
                    exp[1],
                    xerr=exp[3],
                    yerr=exp[2],
                    fmt='.',
                    markersize=0.75,
                    linewidth=0.5)

        ax.plot(q, r, label=l)

    ax.set_yscale('log')
    plt.legend()

    plt.xlabel("Q [nm${}^{-1}$]")
    plt.ylabel("R")

    plt.tight_layout()
    plt.savefig(filename)


def plotSpinAsymmetry(data_pp, data_mm, q, r_pp, r_mm, filename):
    """
    Plot the simulated spin asymmetry as well its 
    experimental counterpart with errorbars
    """

    # compute the errorbars of the spin asymmetry
    delta = numpy.sqrt(4 * (data_pp[1]**2 * data_mm[2]**2 + \
            data_mm[1]**2 * data_pp[2]**2 ) /
                ( data_pp[1] + data_mm[1] )**4 )

    fig = plt.figure()
    ax = fig.add_subplot(111)

    ax.errorbar(data_pp[0],
                (data_pp[1] - data_mm[1])/(data_pp[1] + data_mm[1]),
                xerr=data_pp[3],
                yerr=delta,
                fmt='.',
                markersize=0.75,
                linewidth=0.5)

    ax.plot(q, (r_pp - r_mm)/(r_pp + r_mm))

    plt.gca().set_ylim((-0.3, 0.5))

    plt.xlabel("Q [nm${}^{-1}$]")
    plt.ylabel("Spin asymmetry")

    plt.tight_layout()
    plt.savefig(filename)


####################################################################
#                          Data Handling                           #
####################################################################


def normalizeData(data):
    """
    Removes duplicate q values from the input data,
    normalizes it such that the maximum of the reflectivity is
    unity and rescales the q-axis to inverse nm
    """

    r0 = numpy.where(data[0] - numpy.roll(data[0], 1) == 0)
    data = numpy.delete(data, r0, 1)

    data[0] = data[0]/angstrom
    data[3] = data[3]/angstrom

    norm = numpy.max(data[1])
    data[1] = data[1]/norm
    data[2] = data[2]/norm

    so = numpy.argsort(data[0])
    data = data[:, so]

    return data


def filterData(data, qmin, qmax):
    minIndex = numpy.argmin(numpy.abs(data[0] - qmin))
    maxIndex = numpy.argmin(numpy.abs(data[0] - qmax))

    return data[:, minIndex:maxIndex + 1]


def get_Experimental_data(qmin, qmax):
    if hasattr(get_Experimental_data, "raw_data"):
        data_pp = get_Experimental_data.raw_data_pp
        data_mm = get_Experimental_data.raw_data_mm

    else:
        input_Data = downloadAndExtractData()
        data_pp = normalizeData(input_Data[0])
        data_mm = normalizeData(input_Data[1])

        get_Experimental_data.data_pp = data_pp
        get_Experimental_data.data_mm = data_mm
        get_Experimental_data.raw_data = True

    return (filterData(data_pp, qmin,
                       qmax), filterData(data_mm, qmin, qmax))


def downloadAndExtractData():
    url = "https://www.nist.gov/document/spinelfilmzip"

    if not isfile("spinelfilm.zip"):
        downloadfile = urlopen(url)
        with open("spinelfilm.zip", 'wb') as outfile:
            outfile.write(downloadfile.read())

    zipfile = ZipFile("spinelfilm.zip")
    rawdata = zipfile.open("MAFO_Saturated.refl").\
                  read().decode("utf-8")

    table_pp = match(
        r'.*# "polarization": "\+\+"\n#.*?\n# "units".*?\n(.*?)#.*',
        rawdata, DOTALL).group(1)
    table_mm = match(
        r'.*# "polarization": "\-\-"\n#.*?\n# "units".*?\n(.*?)#.*',
        rawdata, DOTALL).group(1)

    data_pp = numpy.genfromtxt(BytesIO(table_pp.encode()), unpack=True)
    data_mm = numpy.genfromtxt(BytesIO(table_mm.encode()), unpack=True)

    return (data_pp, data_mm)


####################################################################
#                          Main Function                           #
####################################################################

if __name__ == '__main__':

    fixedParams = {
        # parameters from our own fit run
        'q_res': 0.010542945012551425,
        'q_offset': 7.971243487467318e-05,
        'rho_Mafo': 6.370140108715461,
        'rhoM_Mafo': 0.27399566816062926,
        't_Mafo': 137.46913056084736,
        'r_Mao': 8.60487712674644,
        'r_Mafo': 3.7844265311293483
    }

    def run_Simulation_pp(qzs, params):
        return run_simulation(qzs,
                              params,
                              polarization=ba.kvector_t(0, 1, 0),
                              analyzer=ba.kvector_t(0, 1, 0))

    def run_Simulation_mm(qzs, params):
        return run_simulation(qzs,
                              params,
                              polarization=ba.kvector_t(0, -1, 0),
                              analyzer=ba.kvector_t(0, -1, 0))

    qzs = numpy.linspace(qmin, qmax, scan_size)
    q_pp, r_pp = qr(run_Simulation_pp(qzs, fixedParams))
    q_mm, r_mm = qr(run_Simulation_mm(qzs, fixedParams))

    data_pp, data_mm = get_Experimental_data(qmin, qmax)

    plot([q_pp, q_mm], [r_pp, r_mm], [data_pp, data_mm], ["$++$", "$--$"],
         f'MAFO_Saturated.pdf')

    plotSpinAsymmetry(data_pp, data_mm, qzs, r_pp, r_mm,
                      "MAFO_Saturated_spin_asymmetry.pdf")



"""
BornAgain functional test.
Test of rotation/position of core shell particle.
Reference CoreShell particle is compared against the test CoreShell,
which has different dimensions but rotated/translated to be like the reference one.
Particles are placed in the center of middle layer.
"""

from __future__ import print_function
import os, sys, unittest

import utils
from bornagain import *

layer_thickness = 100.0


class TransformCoreShellBoxTest(unittest.TestCase):
    def get_sample(self, particle):
        mAmbience = HomogeneousMaterial("Vacuum", 0, 0)
        mMiddle = HomogeneousMaterial("Teflon", 2.900e-6, 6.019e-9)
        mSubstrate = HomogeneousMaterial("Substrate", 3.212e-6, 3.244e-8)

        layout = ParticleLayout()
        layout.addParticle(particle)

        vacuum_layer = Layer(mAmbience)
        middle_layer = Layer(mMiddle, layer_thickness)
        middle_layer.addLayout(layout)
        substrate = Layer(mSubstrate)

        multi_layer = MultiLayer()
        multi_layer.addLayer(vacuum_layer)
        multi_layer.addLayer(middle_layer)
        multi_layer.addLayer(substrate)
        return multi_layer

    def get_result(self, particle):
        sample = self.get_sample(particle)
        simulation = utils.get_simulation_MiniGISAS(sample)
        simulation.runSimulation()
        return simulation.result()

    def test_SameMaterialCoreShellBox(self):
        """
        Box particle placed in the center of the middle layer is compared against CoreShell particle,
        where core and shell are made of the same material.
        """
        mCore = HomogeneousMaterial("Ag", 1.245e-5, 5.419e-7)

        shell_length = 50.0
        shell_width = 20.0
        shell_height = 10.0
        particle = Particle(
            mCore, FormFactorBox(shell_length, shell_width, shell_height))
        particle.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - shell_height/2))

        reference_data = self.get_result(particle)

        core_length = shell_length/2.0
        core_width = shell_width/2.0
        core_height = shell_height/2.0
        shell = Particle(
            mCore, FormFactorBox(shell_length, shell_width, shell_height))
        core = Particle(
            mCore, FormFactorBox(core_length, core_width, core_height))
        coreshell = ParticleCoreShell(
            shell, core, kvector_t(0, 0, (shell_height - core_height)/2))
        coreshell.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - shell_height/2))

        data = self.get_result(coreshell)

        diff = RelativeDifference(data, reference_data)
        print("test_SameMaterialCoreShell:", diff)
        self.assertLess(diff, 1e-10)

    def test_CoreShellBoxRotateZ(self):
        """
        Two CoreShell particles are compared against each other. Reference CoreShell particle has dimensions (20,50,10)
        and is placed in the center of middle layer. Second CoreShell particle has dimensions (50,20,10) with one
        rotationZ to be like the reference one.
        """
        mCore = HomogeneousMaterial("Ag", 1.245e-5, 5.419e-7)
        mShell = HomogeneousMaterial("AgO2", 8.600e-6, 3.442e-7)

        # building reference CoreShell particle
        shell_length = 20.0
        shell_width = 50.0
        shell_height = 10.0
        core_length = shell_length/2.0
        core_width = shell_width/2.0
        core_height = shell_height/2.0

        core_ref = Particle(
            mCore, FormFactorBox(core_length, core_width, core_height))
        shell_ref = Particle(
            mShell, FormFactorBox(shell_length, shell_width, shell_height))
        coreshell_ref = ParticleCoreShell(
            shell_ref, core_ref,
            kvector_t(0, 0, (shell_height - core_height)/2))
        coreshell_ref.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - shell_height/
                      2))  # center of coreshell in center of the layer

        reference_data = self.get_result(coreshell_ref)

        # building second CoreShell particle
        shell_length = 50.0
        shell_width = 20.0
        shell_height = 10.0
        core_length = shell_length/2.0
        core_width = shell_width/2.0
        core_height = shell_height/2.0

        core = Particle(
            mCore, FormFactorBox(core_length, core_width, core_height))
        shell = Particle(
            mShell, FormFactorBox(shell_length, shell_width, shell_height))
        coreshell = ParticleCoreShell(
            shell, core, kvector_t(0, 0, (shell_height - core_height)/2))
        coreshell.setRotation(RotationZ(90*deg))
        coreshell.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - shell_height/
                      2))  # center of coreshell  in center of the layer

        data = self.get_result(coreshell)

        diff = RelativeDifference(data, reference_data)
        print("test_CoreShellBoxRotateZ:", diff)
        self.assertLess(diff, 1e-10)

    def test_CoreShellBoxRotateY(self):
        """
        Two CoreShell particles are compared against each other. Reference CoreShell particle has dimensions (10,20,50)
        and is placed in the center of middle layer. Second CoreShell particle has dimensions (50,20,10) with one
        rotationZ to be like the reference one.
        """
        mCore = HomogeneousMaterial("Ag", 1.245e-5, 5.419e-7)
        mShell = HomogeneousMaterial("AgO2", 8.600e-6, 3.442e-7)

        # building reference CoreShell particle
        shell_length = 10.0
        shell_width = 20.0
        shell_height = 50.0
        core_length = shell_length/2.0
        core_width = shell_width/2.0
        core_height = shell_height/2.0

        core_ref = Particle(
            mCore, FormFactorBox(core_length, core_width, core_height))
        shell_ref = Particle(
            mShell, FormFactorBox(shell_length, shell_width, shell_height))
        coreshell_ref = ParticleCoreShell(
            shell_ref, core_ref,
            kvector_t(0, 0, (shell_height - core_height)/2))
        coreshell_ref.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - shell_height/
                      2))  # center of coreshell in center of the layer

        reference_data = self.get_result(coreshell_ref)

        # building second CoreShell particle
        shell_length = 50.0
        shell_width = 20.0
        shell_height = 10.0
        core_length = shell_length/2.0
        core_width = shell_width/2.0
        core_height = shell_height/2.0

        core = Particle(
            mCore, FormFactorBox(core_length, core_width, core_height))
        shell = Particle(
            mShell, FormFactorBox(shell_length, shell_width, shell_height))
        coreshell = ParticleCoreShell(
            shell, core, kvector_t(0, 0, (shell_height - core_height)/2))
        coreshell.setRotation(RotationY(90.*deg))
        coreshell.setPosition(
            kvector_t(0, 0, -layer_thickness/
                      2))  # center of coreshell  in center of the layer

        data = self.get_result(coreshell)

        diff = RelativeDifference(data, reference_data)
        print("test_CoreShellBoxRotateY:", diff)
        self.assertLess(diff, 1e-10)

    def test_CoreShellBoxRotateZandY(self):
        """
        Two CoreShell particles are compared against each other. Reference CoreShell particle has dimensions (10,50,20)
        and is placed in the center of middle layer. Second CoreShell particle has dimensions (50,20,10) with two
        rotations to be like the reference one.
        """
        mCore = HomogeneousMaterial("Ag", 1.245e-5, 5.419e-7)
        mShell = HomogeneousMaterial("AgO2", 8.600e-6, 3.442e-7)

        # building reference CoreShell particle
        shell_length = 10.0
        shell_width = 50.0
        shell_height = 20.0
        core_length = shell_length/2.0
        core_width = shell_width/2.0
        core_height = shell_height/2.0

        core_ref = Particle(
            mCore, FormFactorBox(core_length, core_width, core_height))
        shell_ref = Particle(
            mShell, FormFactorBox(shell_length, shell_width, shell_height))
        coreshell_ref = ParticleCoreShell(
            shell_ref, core_ref,
            kvector_t(0, 0, (shell_height - core_height)/2))
        coreshell_ref.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - shell_height/
                      2))  # center of coreshell in center of the layer

        reference_data = self.get_result(coreshell_ref)
        #IntensityDataIOFactory.writeIntensityData(reference_data, "ref_CoreShellBoxRotateZandY.int")

        # building second CoreShell particle
        shell_length = 50.0
        shell_width = 20.0
        shell_height = 10.0
        core_length = shell_length/2.0
        core_width = shell_width/2.0
        core_height = shell_height/2.0

        core = Particle(
            mCore, FormFactorBox(core_length, core_width, core_height))
        shell = Particle(
            mShell, FormFactorBox(shell_length, shell_width, shell_height))
        coreshell = ParticleCoreShell(
            shell, core, kvector_t(0, 0, (shell_height - core_height)/2))
        coreshell.setRotation(RotationZ(90*deg))
        coreshell.rotate(RotationY(90*deg))
        # rotation changes reference point, which now coincide with center of the volume
        coreshell.setPosition(
            kvector_t(0, 0, -layer_thickness/
                      2))  # center of coreshell  in center of the layer

        data = self.get_result(coreshell)

        diff = RelativeDifference(data, reference_data)
        print("test_CoreShellBoxRotateZandY:", diff)
        self.assertLess(diff, 1e-10)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
3 layers system (substrate, teflon, air).
Vacuum layer is populated with spheres with size distribution.
"""
import bornagain as ba
from bornagain import nm


class SampleBuilder:
    """
    Builds complex sample from set of parameters.
    """
    def __init__(self):
        """
        Init SampleBuilder with default sample parameters.
        """
        self.radius = 5.75*ba.nm
        self.sigma = 0.4
        self.distance = 53.6*ba.nm
        self.disorder = 10.5*ba.nm
        self.kappa = 17.5
        self.ptfe_thickness = 22.1*ba.nm
        self.hmdso_thickness = 18.5*ba.nm

    def create_sample(self, params):
        """
        Create sample from given set of parameters.

        Args:
            params: A dictionary containing parameter map.

        Returns:
            A multi layer.
        """
        self.radius = params["radius"]
        self.sigma = params["sigma"]
        self.distance = params["distance"]
        return self.multilayer()

    def multilayer(self):
        """
        Constructs the sample from current parameter values.
        """

        # defining materials
        m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
        m_Si = ba.HomogeneousMaterial("Si", 5.7816e-6, 1.0229e-7)
        m_Ag = ba.HomogeneousMaterial("Ag", 2.2475e-5, 1.6152e-6)
        m_PTFE = ba.HomogeneousMaterial("PTFE", 5.20509e-6, 1.9694e-8)
        m_HMDSO = ba.HomogeneousMaterial("HMDSO", 2.0888e-6, 1.3261e-8)

        # collection of particles with size distribution
        nparticles = 20
        nfwhm = 2.0
        sphere_ff = ba.FormFactorFullSphere(self.radius)

        sphere = ba.Particle(m_Ag, sphere_ff)
        position = ba.kvector_t(0, 0, -1*self.hmdso_thickness)
        sphere.setPosition(position)
        ln_distr = ba.DistributionLogNormal(self.radius, self.sigma)
        par_distr = ba.ParameterDistribution(
            "/Particle/FullSphere/Radius", ln_distr, nparticles, nfwhm,
            ba.RealLimits.limited(0, self.hmdso_thickness/2))
        part_coll = ba.ParticleDistribution(sphere, par_distr)

        # interference function
        interference = ba.InterferenceFunctionRadialParaCrystal(
            self.distance, 1e6*ba.nm)
        interference.setKappa(self.kappa)
        interference.setDomainSize(2e4*nm)
        pdf = ba.FTDistribution1DGauss(self.disorder)
        interference.setProbabilityDistribution(pdf)

        # assembling particle layout
        layout = ba.ParticleLayout()
        layout.addParticle(part_coll, 1)
        layout.setInterferenceFunction(interference)
        layout.setTotalParticleSurfaceDensity(1)

        # roughness
        r_ptfe = ba.LayerRoughness(2.3*ba.nm, 0.3, 5*ba.nm)
        r_hmdso = ba.LayerRoughness(1.1*ba.nm, 0.3, 5*ba.nm)

        # layers
        vacuum_layer = ba.Layer(m_vacuum)
        hmdso_layer = ba.Layer(m_HMDSO, self.hmdso_thickness)
        hmdso_layer.addLayout(layout)
        ptfe_layer = ba.Layer(m_PTFE, self.ptfe_thickness)
        substrate_layer = ba.Layer(m_Si)

        # assembling multilayer
        multi_layer = ba.MultiLayer()
        multi_layer.addLayer(vacuum_layer)
        multi_layer.addLayerWithTopRoughness(hmdso_layer, r_hmdso)
        multi_layer.addLayerWithTopRoughness(ptfe_layer, r_ptfe)
        multi_layer.addLayer(substrate_layer)

        return multi_layer



#!/usr/bin/env python3

import bornagain as ba

def rosenbrock(params):
    x = params["x"].value
    y = params["y"].value
    tmp1 = y - x * x
    tmp2 = 1 - x
    return 100 * tmp1 * tmp1 + tmp2 * tmp2

params = ba.Parameters()
params.add("x", value=-1.2, min=-5.0, max=5.0, step=0.01)
params.add("y", value=1.0, min=-5.0, max=5.0, step=0.01)

minimizer = ba.Minimizer()
result = minimizer.minimize(rosenbrock, params)
print(result.toString())



#!/usr/bin/env python3
"""
Compute radiation trajectories through multilayer sample
for use in figure that explains the Green function.
"""

xd = 52
yd = 10  # detector
s = 3  # source located at -s
T = [6, 6]  # layer thickness
sd = T[0] - s  # distance from source to lower interface
DeltaLayer = [.05, .1]  # refractive index parameter delta, per layer
x0 = 4
xtot = xd - x0  # determine angles so that sum dx matches this

PP = [
    [s, yd],
    [-sd, T[0], yd],
    [s, -T[0], T[0], yd],
    [-sd, T[0], -T[0], T[0], yd],
    [-sd, -T[1], T[1], T[0], yd],
    [-sd, -T[1], T[1], -T[1], T[1], T[0], yd],
    [-sd, -T[1], T[1], T[0], -T[0], T[0], yd],
]

import math, re, scipy.optimize, sys


def inside(y, mi, ma):
    return mi <= y and y <= ma


def xsum(a0, P, D):
    n = len(P)
    if len(D) != n:
        sys.stderr.write("Bug: inconsistent array lengths\n")
        exit(1)
    ret = 0
    for i in range(n):
        a = math.acos(min(.99, math.cos(a0)/(1 - D[i])))
        ret += abs(P[i])/math.tan(a)
    return ret - xtot


cmd0 = ""
cmd0 += "/xd %6.3g def\n" % xd
cmd0 += "/yd %6.3g def\n" % yd
cmd1 = ""
cmd2 = ""
b0 = .8

for P in PP:
    y = -s
    DeltaPath = []
    for dy in P:
        yold = y
        y += dy
        if yold >= 0 and y >= 0:
            DeltaPath.append(0)
        elif inside(y, -T[0], 0) and inside(yold, -T[0], 0):
            DeltaPath.append(DeltaLayer[0])
        elif inside(y, -T[0] - T[1], -T[0]) and inside(
                yold, -T[0] - T[1], -T[0]):
            DeltaPath.append(DeltaLayer[1])
        else:
            sys.stderr.write("Invalid input => layer ill defined\n")
    if y != yd:
        sys.stderr.write("Invalid input => detector not hit\n")
        exit(1)

    a0 = scipy.optimize.newton(xsum, .5, args=(P, DeltaPath))
    print(a0)
    y = -s
    x1 = x0
    x2 = x0
    wgt = 5 - len(P)/2
    cmd1 += "%1i [] lset\n" % (wgt)
    cmd2 += "%1i [] lset\n" % (wgt)
    cmd1 += "%6.3g %6.3g np mv\n" % (x1, y)
    cmd2 += "%6.3g %6.3g np mv\n" % (x2, y)
    for i in range(len(P) - 1):
        a = math.acos(math.cos(a0)/(1 - DeltaPath[i]))
        x1 += abs(P[i])/math.tan(a)
        b = math.acos(math.cos(b0)/(1 - DeltaPath[i]))
        x2 += abs(P[i])/math.tan(b)
        y += P[i]
        cmd1 += "%6.3g %6.3g li\n" % (x1, y)
        cmd2 += "%6.3g %6.3g li\n" % (x2, y)
    cmd1 += "st\n"
    cmd2 += "st\n"
    len1 = math.sqrt((xtot + x0 - x1)**2 + yd**2) - 8
    cmd1 += "%6.3g %6.3g %7.3g .8 %6.3g pfeiL\n" % (x1, y, 180*a0/math.pi,
                                                    len1)
    cmd2 += "%6.3g %6.3g %7.3g .8 %6.3g pfeiL\n" % (x2, y, 180*b0/math.pi,
                                                    7)
    cmd1 += "\n"
    cmd2 += "\n"
#/pfeiL { % (arrow anchored at base) x y rot siz len

t = "% This file is automatically generated by script " + sys.argv[
    0] + "\n% DO NOT EDIT!\n\n"
fd = open("Green1.ps.in", 'r')
t += fd.read()
fd.close

cmd0 += "/b0 %7.3g def\n" % (180*b0/math.pi)
t = re.sub(r'%#0', cmd0, t)
t = re.sub(r'%#1', cmd1, t)
t = re.sub(r'%#2', cmd2, t)

fd = open("Green1.ps", 'w')
fd.write(t)
fd.close



#!/usr/bin/env python3
"""
Fake GISAS data for testing and benchmarking 2D fits.
The sample model basically consists of dilute cylinders on a substrate.
There are, however, some systematic distortions:
  * Refractive indices changed, and absorption strongly enhanced;
  * Cylinders replaced by a mixture of cones and segmented spheroids;
  * Beam wavelength and alpha_i slightly changed.
  * Detector x-axis skewed;
"""

import bornagain as ba
from bornagain import deg, nm
import numpy as np

mat_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
# nominal material constants are (6e-6, 2e-8) and (6e-4, 2e-8)
mat_substrate = ba.HomogeneousMaterial("Substrate", 6.3e-6, 4e-7)
mat_particle = ba.HomogeneousMaterial("Particle", 5.8e-4, 2e-7)


def get_sample(params):
    h = params["cylinder_height"]
    r = params["cylinder_radius"]

    ff1 = ba.FormFactorCone(1.1*r, h, 80*deg)
    ff2 = ba.FormFactorTruncatedSpheroid(r, 3.3*h, 2.5*h/r, 2.2*h)

    layout = ba.ParticleLayout()
    layout.addParticle(ba.Particle(mat_particle, ff1), .4)
    layout.addParticle(ba.Particle(mat_particle, ff2), .6)

    layer_1 = ba.Layer(mat_vacuum)
    layer_1.addLayout(layout)
    layer_2 = ba.Layer(mat_substrate)

    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)
    return sample


def get_simulation(params):
    # nominal beam parameters are 0.1*nm, 0.2*deg
    beam = ba.Beam(10**params['lg(intensity)'], 0.0993*nm,
                   ba.Direction(0.202*deg, 0))

    # nominal detector x-axis starts from -1.5*deg
    det = ba.SphericalDetector(100, -1.49*deg, 1.5*deg, 100, 0, 3*deg)

    sample = get_sample(params)

    simulation = ba.GISASSimulation(beam, sample, det)
    if 'lg(background)' in params:
        simulation.setBackground(
            ba.ConstantBackground(10**params['lg(background)']))

    return simulation


def model_parameters():
    return {
        'lg(intensity)': 6,
        'lg(background)': 0.5,
        'cylinder_height': 4*nm,
        'cylinder_radius': 5*nm,
    }


def fake_data():
    """
    Generate fake "experimental" data, and save them as numpy array.
    """

    params = model_parameters()

    # Compute model distribution
    simulation = get_simulation(params)
    simulation.runSimulation()
    theory = simulation.result().array()

    # Draw noisy data
    np.random.seed(0)
    data = np.random.poisson(theory)

    # Save to numpy
    np.savetxt("faked-gisas1.txt.gz", data)


if __name__ == '__main__':
    fake_data()



# this script is used to manually tweak swig-generated wrapper

import sys
import os
import string


# compatibility with windows filename conventions
def windowsify(filename):
    if (sys.platform == "win32"):
        new_filename = filename.replace("/", "\\")
    else:
        new_filename = filename

    return os.path.abspath(new_filename)


if len(sys.argv) != 3:
    print("Usage: tweaks.py inputfile outputfile")
    exit(1)

in_name = windowsify(sys.argv[1])
out_name = windowsify(sys.argv[2])

print("tweaks.py:", in_name, out_name)

fd = open(in_name, "r")
text_in = fd.read().splitlines()
fd.close()

text_out = []
for line in text_in:
    line = line.replace("_libBornAgainBase.ICloneable_transferToCPP(self)",
                        "self.__disown__()")
    line = line.replace("_libBornAgainCore.Histogram1D_binCenters(self)",
                        "self.binCentersNumpy()")
    line = line.replace("_libBornAgainCore.Histogram1D_binErrors(self)",
                        "self.binErrorsNumpy()")
    line = line.replace("_libBornAgainCore.Histogram1D_binValues(self)",
                        "self.binValuesNumpy()")
    text_out.append(line)

# done the search/replace
fd = open(out_name, "w")
fd.write("\n".join(text_out))
fd.close()



#!/usr/bin/env python3
"""
Basic example of specular reflectometry simulation with BornAgain.
The sample consists of 20 alternating Ti and Ni layers.
"""
import bornagain as ba
from bornagain import deg, angstrom


def get_sample():
    # Define materials
    m_ambient = ba.MaterialBySLD("Vacuum", 0, 0)
    m_ti = ba.MaterialBySLD("Ti", -1.9493e-06, 0)
    m_ni = ba.MaterialBySLD("Ni", 9.4245e-06, 0)
    m_substrate = ba.MaterialBySLD("SiSubstrate", 2.0704e-06, 0)

    # Define layers
    ambient_layer = ba.Layer(m_ambient)
    ti_layer = ba.Layer(m_ti, 30*angstrom)
    ni_layer = ba.Layer(m_ni, 70*angstrom)
    substrate_layer = ba.Layer(m_substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(ambient_layer)
    for i in range(10):
        sample.addLayer(ti_layer)
        sample.addLayer(ni_layer)
    sample.addLayer(substrate_layer)

    return sample


def get_simulation(sample, scan_size=500):
    simulation = ba.SpecularSimulation()
    scan = ba.AngularSpecScan(1.54*angstrom, scan_size, 0, 2*deg)
    simulation.setScan(scan)
    simulation.setSample(sample)
    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)



#!/usr/bin/env python3
"""
Square lattice of cylinders inside finite layer with usage of average material
"""
import bornagain as ba
from bornagain import deg, nm, kvector_t


def get_sample(cyl_height=5*nm):
    """
    Returns a sample with cylinders on a substrate.
    """
    # defining materials
    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_layer = ba.HomogeneousMaterial("Layer", 3e-6, 2e-8)
    m_substrate = ba.HomogeneousMaterial("Substrate", 6e-6, 2e-8)
    m_particle = ba.HomogeneousMaterial("Particle", 3e-5, 2e-8)

    # cylindrical particle
    cylinder_ff = ba.FormFactorCylinder(5*nm, cyl_height)
    cylinder = ba.Particle(m_particle, cylinder_ff)
    position = ba.kvector_t(0, 0, -cyl_height)
    particle_layout = ba.ParticleLayout()
    particle_layout.addParticle(cylinder, 1, position)

    # interference function
    interference = ba.InterferenceFunction2DLattice(
        ba.SquareLattice2D(15*nm, 0))
    pdf = ba.FTDecayFunction2DCauchy(300*nm, 300*nm, 0)
    interference.setDecayFunction(pdf)
    particle_layout.setInterferenceFunction(interference)

    vacuum_layer = ba.Layer(m_vacuum)
    intermediate_layer = ba.Layer(m_layer, 5*nm)
    intermediate_layer.addLayout(particle_layout)
    substrate_layer = ba.Layer(m_substrate)

    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    multi_layer.addLayer(intermediate_layer)
    multi_layer.addLayer(substrate_layer)
    return multi_layer


def get_simulation(sample):
    beam = ba.Beam(1, 0.1*nm, ba.Direction(0.2*deg, 0))
    detector = ba.SphericalDetector(100, -2*deg, 2*deg, 100, 0, 2*deg)
    simulation = ba.GISASSimulation(beam, sample, detector)
    simulation.getOptions().setUseAvgMaterials(True)
    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)



#!/usr/bin/env python3
"""
Extended example for simulation results treatment (cropping, slicing, exporting)
The standard "Cylinders in DWBA" sample is used to setup the simulation.
"""
import math
import random
import bornagain as ba
from bornagain import angstrom, deg, nm, nm2, kvector_t
import ba_plot
from matplotlib import pyplot as plt
from matplotlib import rcParams


def get_sample():
    """
    Returns a sample with uncorrelated cylinders on a substrate.
    """

    # Define materials
    material_Particle = ba.HomogeneousMaterial("Particle", 0.0006, 2e-08)
    material_Substrate = ba.HomogeneousMaterial("Substrate", 6e-06, 2e-08)
    material_Vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)

    # Define form factors
    ff = ba.FormFactorCylinder(5*nm, 5*nm)

    # Define particles
    particle = ba.Particle(material_Particle, ff)

    # Define particle layouts
    layout = ba.ParticleLayout()
    layout.addParticle(particle)
    layout.setTotalParticleSurfaceDensity(0.01)

    # Define layers
    layer_1 = ba.Layer(material_Vacuum)
    layer_1.addLayout(layout)
    layer_2 = ba.Layer(material_Substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def get_simulation(sample):
    """
    Returns a GISAXS simulation with beam and detector defined.
    """
    beam = ba.Beam(1e5, 1*angstrom, ba.Direction(0.2*deg, 0))
    det = ba.SphericalDetector(201, -2*deg, 2*deg, 201, 0, 2*deg)
    simulation = ba.GISASSimulation(beam, sample, det)
    return simulation


def get_noisy_image(hist):
    """
    Returns clone of input histogram filled with additional noise
    """
    result = hist.clone()
    noise_factor = 2.0
    for i in range(0, result.getTotalNumberOfBins()):
        amplitude = result.binContent(i)
        sigma = noise_factor*math.sqrt(amplitude)
        noisy_amplitude = random.gauss(amplitude, sigma)
        result.setBinContent(i, noisy_amplitude)
    return result


def plot_histogram(hist, **kwargs):
    ba.plot_histogram(hist,
                      xlabel=r'$\varphi_f ^{\circ}$',
                      ylabel=r'$\alpha_f ^{\circ}$',
                      zlabel="",
                      **kwargs)


def get_relative_difference(hist):
    """
    Creates noisy histogram made of original histogram,
    then creates and plots a relative difference histogram: (noisy-hist)/hist
    """
    noisy = get_noisy_image(hist)
    return noisy.relativeDifferenceHistogram(hist)


def plot_slices(hist):
    """
    Plot 1D slices along y-axis at certain x-axis values.
    """
    noisy = get_noisy_image(hist)

    # projection along Y, slice at fixed x-value
    proj1 = noisy.projectionY(0)
    plt.semilogy(proj1.binCenters(),
                 proj1.binValues(),
                 label=r'$\phi=0.0^{\circ}$')

    # projection along Y, slice at fixed x-value
    proj2 = noisy.projectionY(0.5)  # slice at fixed value
    plt.semilogy(proj2.binCenters(),
                 proj2.binValues(),
                 label=r'$\phi=0.5^{\circ}$')

    # projection along Y for all X values between [xlow, xup], averaged
    proj3 = noisy.projectionY(0.41, 0.59)
    plt.semilogy(proj3.binCenters(),
                 proj3.array(ba.IHistogram.AVERAGE),
                 label=r'$<\phi>=0.5^{\circ}$')

    plt.xlim(proj1.getXmin(), proj1.getXmax())
    plt.ylim(proj2.getMinimum(), proj1.getMaximum()*10)
    plt.xlabel(r'$\alpha_f ^{\circ}$', fontsize=16)
    plt.legend(loc='upper right')
    plt.tight_layout()


def plot(hist):
    """
    Runs different plotting functions one by one
    to demonstrate trivial data presentation tasks.
    """
    plt.figure(figsize=(12.80, 10.24))

    plt.subplot(2, 2, 1)
    ba_plot.plot_histogram(hist)
    plt.title("Intensity as colormap")

    plt.subplot(2, 2, 2)
    crop = hist.crop(-1, 0.5, 1, 1)
    ba_plot.plot_histogram(crop)
    plt.title("Cropping")

    plt.subplot(2, 2, 3)
    reldiff = get_relative_difference(hist)
    ba_plot.plot_histogram(reldiff, intensity_min=1e-03, intensity_max=10)
    plt.title("Relative difference")

    plt.subplot(2, 2, 4)
    plot_slices(hist)
    plt.title("Various slicing of 2D into 1D")

    # save to file
    # result.save("result.int")
    # result.save("result.tif")
    # result.save("result.txt")
    # result.save("result.int.gz")
    # result.save("result.tif.gz")
    # result.save("result.txt.gz")
    # result.save("result.int.bz2")
    # result.save("result.tif.bz2")
    # result.save("result.txt.bz2")

    plt.tight_layout()
    plt.show()


def simulate_and_plot():
    sample = get_sample()
    simulation = get_simulation(sample)
    simulation.runSimulation()
    hist = simulation.result().histogram2d()
    plot(hist)


if __name__ == '__main__':
    simulate_and_plot()



"""
Process gitlog and create a file with number of lines of code.
"""
from datetime import datetime
from .file_types import FileTypes, filetype
import re
from email.utils import parsedate
import subprocess


def gitlog():
    """
    Execute gitlog command and make generator over lines in the log
    """
    p = subprocess.Popen(['git', 'log', 'develop', '--reverse', '-p'], stdout=subprocess.PIPE)
    for line in iter(p.stdout.readline, b''):
        decoded = line.decode('latin1')
        if decoded and len(decoded):
            yield decoded.strip()


class Commit:
    """
    Contains commit info and accumulated number of lines of code (file type dependent).
    """
    def __init__(self):
        self.date = datetime.today()
        self.added_lines = 0
        self.removed_lines = 0
        self.loc_for_type = FileTypes.loc_for_type()
        self.hsh = None
        self.who = None
        self.cmt = None

    def increment_loc(self, file_type):
        self.loc_for_type[file_type] += 1
        self.added_lines += 1

    def decrement_loc(self, file_type):
        self.loc_for_type[file_type] -= 1
        self.removed_lines += 1


class DayHistory:
    """
    Number of lines added or deleted for given day.
    """
    def __init__(self, date, locs):
        self.date = date
        self.loc_for_type = locs


class HistoryCollector:
    def __init__(self):
        self.last_commit = Commit()
        self.data = []
        self.locs = 0
        self.fc = 0
        self.file_type_ppp = FileTypes.UNDEF
        self.file_type_mmm = FileTypes.UNDEF
        self.start_date = datetime(2012, 4, 1)
        self.days_history = {}  # DayHistory vs. number of days since beginning of coding

    def pop(self):
        if not self.last_commit.added_lines:
            return
        pstr="%s %8u %5s %5s %7s %s"%(self.last_commit.date, self.locs, '+' + str(self.last_commit.added_lines), '-' + str(self.last_commit.removed_lines), self.last_commit.hsh, self.last_commit.who)
        print(pstr)
        delta = (self.last_commit.date - self.start_date).days
        self.days_history[delta] = DayHistory(self.last_commit.date, self.last_commit.loc_for_type)

        tmp = list(self.last_commit.loc_for_type)
        self.last_commit = Commit()
        self.last_commit.loc_for_type = tmp

    def run(self):
        nnn = 0
        for x in gitlog():
            nnn += 1
            if x.startswith('commit'):
                self.pop()
                self.last_commit.hsh = x[7:14]

            if x.startswith('Author'):
                self.last_commit.who = x.replace("Author: ", '').replace('\n', '')
                self.last_commit.who = re.sub(">.*", "", self.last_commit.who)
                self.last_commit.who = re.sub(".*<", "", self.last_commit.who)

            if x.startswith('Date'):
                self.fc = 1
                self.last_commit.date = datetime(*parsedate(x[5:])[:7])

            if self.fc == 2:
                self.last_commit.cmt = x[:-1]
                self.fc = 0

            if self.fc == 1:
                if len(x) == 1:
                    self.fc = 2

            if x.startswith('+++'):
                self.file_type_ppp = filetype(x)

            if x.startswith('---'):
                self.file_type_mmm = filetype(x)

            if x.startswith('+') and not x.startswith('+++'):
                self.last_commit.increment_loc(self.file_type_ppp)
                if self.file_type_ppp <FileTypes.PYAPI:
                    self.locs += 1

            if x.startswith('-') and not x.startswith('---'):
                self.last_commit.decrement_loc(self.file_type_mmm)
                if self.file_type_mmm < FileTypes.PYAPI:
                    self.locs -= 1

            # if nnn>1000000:
            #     break

        self.pop()

    def save_report(self, filename):
        print("Saving report in {0}".format(filename))
        with open(filename, 'w') as the_file:
            for key in self.days_history:
                dayhist = self.days_history[key]
                pstr = "%s %s \n" % (str(dayhist.date), ' '.join(str(e) for e in dayhist.loc_for_type))
                the_file.write(pstr)



#!/usr/bin/env python3
"""
Run plotting code in batch mode: generate figure, but don't plot to terminal.

This script is used in BornAgain:
- with option -s in PyExamples tests, to check functionality of Python examples;
- with option -l to remake full-size images.
"""

import matplotlib, os, re, sys
matplotlib.use('Agg')
from matplotlib import pyplot as plt


def exec_full(script, filename):
    """
    Executes embedded python script.
    http://stackoverflow.com/questions/436198/what-is-an-alternative-to-execfile-in-python-3
    """
    import os
    global_namespace = {
        "__file__": filename,
        "__name__": "__main__",
        "__no_terminal__": True
    }
    sys.argv = [filename, 'noshow']
    exec(compile(script, filename, 'exec'), global_namespace)

def reduce_nbin(t):
    """
    Overwrites script lines that set nbin, nx, ny
    """
#    pat = re.compile(r'(^\s+(nbin|nx|ny) = )(\d+)$')
#    ret = []
#    for l in t.split('\n'):
#        m = re.match(pat, l)
#        if m:
#            oldsize = int(m.group(3))
#            newsize = min(10, oldsize)
#            lout = re.sub(pat, m.group(1)+f'{newsize}', l)
#        else:
#            lout = l
#        ret.append(lout)
#    return '\n'.join(ret)
    return t

def run_example(mode_short, filename, output_dir):
    """
    Tries to run python example and produce a *.png image
    """
    # Read script from file.
    if not os.path.exists(filename):
        raise Exception("Example script '" + filename + "' not found")
    print("Input script: " + filename, flush=True)
    with open(filename, 'r') as file:
        script = file.read()

    # Detect or impose figure size.
    m = re.search(r'plt\.figure\(.+?figsize=\((.+?),(.+?)\)', script)
    if m: # set figure size as in script
        figsize = (float(m.group(1)), float(m.group(2)))
    else: # script does not specify figure size
        figsize = (640/72, 480/72)
    fig = plt.figure(figsize=(figsize[0], figsize[1]))

    if mode_short:
        script = reduce_nbin(script)

    # Run modified script.
    exec_full(script, filename)
    print("Input script completed", flush=True)

    # Generate output figure.
    plot_file_name = os.path.join(
        output_dir,
        os.path.splitext(os.path.basename(filename))[0] + ".png")
    print("Output image: " + plot_file_name)
    plt.savefig(plot_file_name, bbox_inches='tight')
    plt.close(fig)

    # Check obtained figure.
    imgSize = os.path.getsize(plot_file_name)
    if imgSize == 0:
        raise Exception("Image file is empty")
    if imgSize < 20000:
        raise Exception("Image file is unplausibly small")


if __name__ == '__main__':

    if len(sys.argv) != 4:
        exit("Usage: check_functionality -s|-f <script_to_test> <output_dir>")
    if sys.argv[1]=='-s':
        mode_short = True
    elif sys.argv[1]=='-f':
        mode_short = False
    else:
        exit("Use flag -s|-f for short or long runs")

    run_example(mode_short, sys.argv[2], sys.argv[3])



# Functional test: Magnetic cylinders in DWBA with zero magnetic field
from __future__ import print_function
import sys
import os
import numpy
import gzip
from utils import get_difference, REFERENCE_DIR

from bornagain import *


# ----------------------------------
# describe sample and run simulation
# ----------------------------------
def getSimulationIntensity(rho_beam, efficiency):
    print("- simulate", flush=True)
    # defining materials
    mAmbience = HomogeneousMaterial("Vacuum", 0, 0)
    mSubstrate = HomogeneousMaterial("Substrate", 15e-6, 0)

    magnetization = kvector_t(0, 1e6, 0)

    magParticle = HomogeneousMaterial("magParticle", 5e-6, 0,
                                      magnetization)
    # collection of particles
    cylinder_ff = FormFactorCylinder(5*nm, 5*nm)
    cylinder = Particle(magParticle, cylinder_ff)

    particle_layout = ParticleLayout()
    particle_layout.addParticle(cylinder, 1)
    interference = InterferenceFunctionNone()
    particle_layout.setInterferenceFunction(interference)

    # vacuum layer with particles and substrate form multi layer
    vacuum_layer = Layer(mAmbience)
    vacuum_layer.addLayout(particle_layout)
    substrate_layer = Layer(mSubstrate, 0)
    multi_layer = MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    multi_layer.addLayer(substrate_layer)

    # build and run experiment
    simulation = GISASSimulation()
    simulation.setDetectorParameters(100, -1*deg, 1*deg, 100, 0, 2*deg)
    zplus = kvector_t(0, 0, 1)
    simulation.detector().setAnalyzerProperties(zplus, efficiency, 0.5)
    simulation.setBeamParameters(1*angstrom, 0.2*deg, 0)
    simulation.beam().setPolarization(rho_beam)
    simulation.setSample(multi_layer)
    simulation.beam().setIntensity(1e9)
    simulation.runSimulation()
    print("- - simulation done", flush=True)
    return simulation.result()


# TODO: revert to some standard solution. For other tests, get_reference_data comes from utils.py.
def get_reference_data(filename):
    """
    read and return reference data from file
    """
    path = os.path.join(REFERENCE_DIR, filename)
    print("- read reference from", path, flush=True)
    ret = IntensityDataIOFactory.readIntensityData(path)
    print("- - reference read", flush=True)
    return ret


# --------------------------------------------------------------
# run test and analyse test results
# --------------------------------------------------------------
def run_test():
    print("run test", flush=True)
    zplus = kvector_t(0, 0, 1)
    zmin = kvector_t(0, 0, -1)

    # IntensityDataIOFactory.writeIntensityData(getSimulationIntensity(zplus, 1), 'polmagcylinders2_reference_00.int')
    # IntensityDataIOFactory.writeIntensityData(getSimulationIntensity(zplus, -1), 'polmagcylinders2_reference_01.int')
    # IntensityDataIOFactory.writeIntensityData(getSimulationIntensity(zmin, 1), 'polmagcylinders2_reference_10.int')
    # IntensityDataIOFactory.writeIntensityData(getSimulationIntensity(zmin, -1), 'polmagcylinders2_reference_11.int')
    diff = 0.0
    diff += get_difference(
        getSimulationIntensity(zplus, 1).array(),
        get_reference_data(
            'polmagcylinders2_reference_00.int.gz').getArray())
    diff += get_difference(
        getSimulationIntensity(zplus, -1).array(),
        get_reference_data(
            'polmagcylinders2_reference_01.int.gz').getArray())
    diff += get_difference(
        getSimulationIntensity(zmin, 1).array(),
        get_reference_data(
            'polmagcylinders2_reference_10.int.gz').getArray())
    diff += get_difference(
        getSimulationIntensity(zmin, -1).array(),
        get_reference_data(
            'polmagcylinders2_reference_11.int.gz').getArray())

    diff /= 4.0
    status = "OK"
    if diff > 2e-10:
        status = "FAILED"
    return "PolarizedDWBAZeroMag", "functional test: polarized DWBA with non-zero magnetic field", diff, status


if __name__ == '__main__':
    name, description, diff, status = run_test()
    print(name, description, diff, status)
    if ("FAILED" in status):
        exit(1)



#!/usr/bin/env python3
"""
This example demonstrates how to fit a complex experimental setup using BornAgain.
It is based on real data published in  https://doi.org/10.1002/advs.201700856
by A. Glavic et al.
In this example we utilize the scalar reflectometry  engine to fit polarized
data without spin-flip for performance reasons.
"""

import sys
import numpy
import matplotlib.pyplot as plt

from scipy.optimize import differential_evolution

import bornagain as ba
from bornagain import angstrom

from zipfile import ZipFile

# number of points on which the computed result is plotted
scan_size = 1500

# restrict the Q-range of the data used for fitting
qmin = 0.08
qmax = 1.4

####################################################################
#                  Create Sample and Simulation                    #
####################################################################


def get_sample(parameters, sign, ms150=1):

    m_Air = ba.MaterialBySLD("Air", 0, 0)
    m_PyOx  = ba.MaterialBySLD("PyOx",
                               (parameters["sld_PyOx_real"] + \
                                 sign * ms150 * parameters["msld_PyOx"] )* 1e-6,
                               parameters["sld_PyOx_imag"] * 1e-6)
    m_Py2   = ba.MaterialBySLD("Py2",
                               ( parameters["sld_Py2_real"] + \
                                 sign * ms150 * parameters["msld_Py2"] ) * 1e-6,
                               parameters["sld_Py2_imag"] * 1e-6)
    m_Py1   = ba.MaterialBySLD("Py1",
                               ( parameters["sld_Py1_real"] + \
                                 sign * ms150 * parameters["msld_Py1"] ) * 1e-6,
                               parameters["sld_Py1_imag"] * 1e-6)
    m_SiO2 = ba.MaterialBySLD("SiO2", parameters["sld_SiO2_real"]*1e-6,
                              parameters["sld_SiO2_imag"]*1e-6)
    m_Si = ba.MaterialBySLD("Substrate", parameters["sld_Si_real"]*1e-6,
                            parameters["sld_Si_imag"]*1e-6)

    l_Air = ba.Layer(m_Air)
    l_PyOx = ba.Layer(m_PyOx, parameters["t_PyOx"]*angstrom)
    l_Py2 = ba.Layer(m_Py2, parameters["t_Py2"]*angstrom)
    l_Py1 = ba.Layer(m_Py1, parameters["t_Py1"]*angstrom)
    l_SiO2 = ba.Layer(m_SiO2, parameters["t_SiO2"]*angstrom)
    l_Si = ba.Layer(m_Si)

    r_PyOx = ba.LayerRoughness()
    r_Py2 = ba.LayerRoughness()
    r_Py1 = ba.LayerRoughness()
    r_SiO2 = ba.LayerRoughness()
    r_Si = ba.LayerRoughness()

    r_PyOx.setSigma(parameters["r_PyOx"]*angstrom)
    r_Py2.setSigma(parameters["r_Py2"]*angstrom)
    r_Py1.setSigma(parameters["r_Py1"]*angstrom)
    r_SiO2.setSigma(parameters["r_SiO2"]*angstrom)
    r_Si.setSigma(parameters["r_Si"]*angstrom)

    sample = ba.MultiLayer()

    sample.addLayer(l_Air)
    sample.addLayerWithTopRoughness(l_PyOx, r_PyOx)
    sample.addLayerWithTopRoughness(l_Py2, r_Py2)
    sample.addLayerWithTopRoughness(l_Py1, r_Py1)
    sample.addLayerWithTopRoughness(l_SiO2, r_SiO2)
    sample.addLayerWithTopRoughness(l_Si, r_Si)

    sample.setRoughnessModel(ba.RoughnessModel.NEVOT_CROCE)

    return sample


def get_simulation(q_axis, fitParams, sign, ms150=False):

    n_points = 25
    n_sig = 3.0

    parameters = dict(fitParams, **fixedParams)

    q_distr = ba.RangedDistributionGaussian(n_points, n_sig)

    dq = parameters["dq"]*q_axis
    scan = ba.QSpecScan(q_axis)
    scan.setAbsoluteQResolution(q_distr, dq)

    simulation = ba.SpecularSimulation()
    simulation.setScan(scan)
    simulation.beam().setIntensity(parameters["intensity"])

    if ms150:
        sample = get_sample(parameters=parameters,
                            sign=sign,
                            ms150=parameters["ms150"])
    else:
        sample = get_sample(parameters=parameters, sign=sign, ms150=1)

    simulation.setSample(sample)
    simulation.setBackground(ba.ConstantBackground(5e-7))

    return simulation


def run_simulation(q_axis, fitParams, *, sign, ms150=False):
    simulation = get_simulation(q_axis, fitParams, sign, ms150)
    simulation.runSimulation()

    return simulation


def qr(result):
    """
    Returns two arrays that hold the q-values as well as the
    reflectivity from a given simulation result
    """
    q = numpy.array(result.result().axis(ba.Axes.QSPACE))
    r = numpy.array(result.result().array(ba.Axes.QSPACE))

    return q, r


####################################################################
#                         Plot Handling                            #
####################################################################


def plot(qs, rs, exps, shifts, labels, filename):
    """
    Plot the simulated result together with the experimental data
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)

    for q, r, exp, shift, l in zip(qs, rs, exps, shifts, labels):

        ax.errorbar(exp[0],
                    exp[1]/shift,
                    yerr=exp[2]/shift,
                    fmt='.',
                    markersize=0.75,
                    linewidth=0.5)

        ax.plot(q, r/shift, label=l)

    ax.set_yscale('log')
    plt.legend()

    plt.xlabel("Q [nm${}^{-1}$]")
    plt.ylabel("R")

    plt.tight_layout()
    plt.savefig(filename)


def plot_sld_profile(fitParams, filename):

    plt.figure()
    parameters = dict(fitParams, **fixedParams)

    z_300_p, sld_300_p = ba.materialProfile(get_sample(parameters, 1))
    z_300_m, sld_300_m = ba.materialProfile(get_sample(parameters, -1))

    z_150_p, sld_150_p = ba.materialProfile(
        get_sample(parameters, 1, ms150=parameters["ms150"]))
    z_150_m, sld_150_m = ba.materialProfile(
        get_sample(parameters, -1, ms150=parameters["ms150"]))

    plt.figure()
    nsld = plt.plot(z_300_p, numpy.real(sld_300_p)*1e6, label=r"300K $+$")
    nsld = plt.plot(z_300_m, numpy.real(sld_300_m)*1e6, label=r"300K $-$")

    nsld = plt.plot(z_150_p, numpy.real(sld_150_p)*1e6, label=r"150K $+$")
    nsld = plt.plot(z_150_m, numpy.real(sld_150_m)*1e6, label=r"150K $-$")

    plt.xlabel(r"$z$ [A]")
    plt.ylabel(r"$\delta(z) \cdot 10^6$")

    plt.legend()
    plt.tight_layout()
    plt.savefig(filename)
    plt.close()


####################################################################
#                          Data Handling                           #
####################################################################


def normalizeData(data):
    """
    Removes duplicate q values from the input data,
    normalizes it such that the maximum of the reflectivity is
    unity and rescales the q-axis to inverse nm
    """
    # delete repeated data
    r0 = numpy.where(data[0] - numpy.roll(data[0], 1) == 0)
    data = numpy.delete(data, r0, 1)

    data[0] = data[0]/angstrom
    norm = numpy.max(data[1])
    data[1] = data[1]/norm
    data[2] = data[2]/norm

    # sort by q axis
    so = numpy.argsort(data[0])
    data = data[:, so]

    return data


def get_Experimental_data(filename, qmin, qmax):

    if not hasattr(get_Experimental_data, "zipfile"):
        get_Experimental_data.zipfile = ZipFile("honeycomb_data.zip")

    input_Data = numpy.genfromtxt(
        get_Experimental_data.zipfile.open(filename),
        unpack=True,
        usecols=(0, 2, 3))
    data = normalizeData(input_Data)

    minIndex = numpy.argmin(numpy.abs(data[0] - qmin))
    maxIndex = numpy.argmin(numpy.abs(data[0] - qmax))

    return data[:, minIndex:maxIndex + 1]


####################################################################
#                          Fit Function                            #
####################################################################


def relative_difference(sim, exp):
    result = (exp - sim)/(exp + sim)
    return numpy.sum(result*result)/len(sim)


def create_Parameter_dictionary(parameterNames, *args):
    return {name: value for name, value in zip(parameterNames, *args)}


class FitObjective:
    def __init__(self, q_axis, rdata, simulationFactory, parameterNames):
        if isinstance(q_axis, list) and isinstance(rdata, list) and \
                                    isinstance(simulationFactory, list):
            self._q = q_axis
            self._r = rdata
            self._simulationFactory = simulationFactory

        elif not isinstance(q_axis, list) and not isinstance(rdata, list) \
                              and not isinstance(simulationFactory, list):
            self._q = [q_axis]
            self._r = [rdata]
            self._simulationFactory = [simulationFactory]

        else:
            raise Exception("Inconsistent parameters")

        self._parameterNames = parameterNames

    def __call__(self, *args):
        fitParameters = create_Parameter_dictionary(self._parameterNames,
                                                    *args)
        print(f"FitParamters = {fitParameters}")

        result_metric = 0

        for q, r, sim in zip(self._q, self._r, self._simulationFactory):
            sim_result = sim(q, fitParameters).result().array()
            result_metric += relative_difference(sim_result, r)

        return result_metric


def run_fit_differential_evolution(q_axis, rdata, simulationFactory,
                                   startParams):

    bounds = [(par[1], par[2]) for n, par in startParams.items()]
    parameters = [par[0] for n, par in startParams.items()]
    parameterNames = [n for n, par in startParams.items()]
    print(f"Bounds = {bounds}")

    objective = FitObjective(q_axis, rdata, simulationFactory,
                             parameterNames)

    chi2_initial = objective(parameters)

    result = differential_evolution(objective,
                                    bounds,
                                    maxiter=200,
                                    popsize=len(bounds)*10,
                                    mutation=(0.5, 1.5),
                                    disp=True,
                                    tol=1e-2)

    resultParameters = create_Parameter_dictionary(parameterNames,
                                                   result.x)
    chi2_final = objective(resultParameters.values())

    print(f"Initial chi2: {chi2_initial}")
    print(f"Final chi2: {chi2_final}")
    return resultParameters


####################################################################
#                          Main Function                           #
####################################################################

if __name__ == '__main__':

    fixedParams = {
        "sld_PyOx_imag": (0, 0, 0),
        "sld_Py2_imag": (0, 0, 0),
        "sld_Py1_imag": (0, 0, 0),
        "sld_SiO2_imag": (0, 0, 0),
        "sld_Si_imag": (0, 0, 0),
        "sld_SiO2_real": (3.47, 3, 4),
        "sld_Si_real": (2.0704, 2, 3),
        "dq": (0.018, 0, 0.1),
    }

    if len(sys.argv) > 1 and sys.argv[1] == "fit":

        # some sensible start parameters for fitting
        startParams = {
            "intensity": (1.04, 0, 3),
            "t_PyOx": (77, 60, 100),
            "t_Py2": (56, 40, 70),
            "t_Py1": (64, 50, 80),
            "t_SiO2": (16, 10, 30),
            "sld_PyOx_real": (1.915, 1.6, 2.2),
            "sld_Py2_real": (5, 3, 6),
            "sld_Py1_real": (4.62, 3, 6),
            "r_PyOx": (27, 5, 35),
            "r_Py2": (12, 5, 20),
            "r_Py1": (12, 5, 20),
            "r_SiO2": (17, 2, 25),
            "r_Si": (18, 2, 25),
            "msld_PyOx": (0.25, 0, 1),
            "msld_Py2": (0.63, 0, 1),
            "msld_Py1": (0.64, 0, 1),
            "ms150": (1, 0.9, 1.1),
        }

        fit = True

    else:
        # result from our own fitting
        startParams = {
            'intensity': 0.9482344993285265,
            't_PyOx': 74.97056190221168,
            't_Py2': 61.75823766477464,
            't_Py1': 54.058310970786316,
            't_SiO2': 23.127048588278402,
            'sld_PyOx_real': 2.199791584033569,
            'sld_Py2_real': 4.980316982224387,
            'sld_Py1_real': 4.612135848532186,
            'r_PyOx': 31.323366207013787,
            'r_Py2': 9.083768897940645,
            'r_Py1': 5,
            'r_SiO2': 14.43455709065263,
            'r_Si': 14.948233893986075,
            'msld_PyOx': 0.292684104601585,
            'msld_Py2': 0.5979447434271339,
            'msld_Py1': 0.56376339230351,
            'ms150': 1.083311702077648
        }

        startParams = {d: (v, ) for d, v in startParams.items()}
        fit = False

    fixedParams = {d: v[0] for d, v in fixedParams.items()}
    paramsInitial = {d: v[0] for d, v in startParams.items()}

    def run_Simulation_300_p(qzs, params):
        return run_simulation(qzs, params, sign=1)

    def run_Simulation_300_m(qzs, params):
        return run_simulation(qzs, params, sign=-1)

    def run_Simulation_150_p(qzs, params):
        return run_simulation(qzs, params, sign=1, ms150=True)

    def run_Simulation_150_m(qzs, params):
        return run_simulation(qzs, params, sign=-1, ms150=True)

    qzs = numpy.linspace(qmin, qmax, scan_size)
    q_300_p, r_300_p = qr(run_Simulation_300_p(qzs, paramsInitial))
    q_300_m, r_300_m = qr(run_Simulation_300_m(qzs, paramsInitial))

    q_150_p, r_150_p = qr(run_Simulation_150_p(qzs, paramsInitial))
    q_150_m, r_150_m = qr(run_Simulation_150_m(qzs, paramsInitial))

    data_300_p = get_Experimental_data("honeycomb_300_p.dat", qmin, qmax)
    data_300_m = get_Experimental_data("honeycomb_300_m.dat", qmin, qmax)

    data_150_p = get_Experimental_data("honeycomb_150_p.dat", qmin, qmax)
    data_150_m = get_Experimental_data("honeycomb_150_m.dat", qmin, qmax)

    plot_sld_profile(paramsInitial,
                     f"Honeycomb_Fit_sld_profile_initial.pdf")
    plot([q_300_p, q_300_m, q_150_p, q_150_m],
         [r_300_p, r_300_m, r_150_p, r_150_m],
         [data_300_p, data_300_m, data_150_p, data_150_m], [1, 1, 10, 10],
         ["300K $+$", "300K $-$", "150K $+$", "150K $-$"],
         f"Honeycomb_Fit_reflectivity_initial.pdf")

    # fit and plot fit
    if fit:
        dataSimTuple = [[
            data_300_p[0], data_300_m[0], data_150_p[0], data_150_m[0]
        ], [data_300_p[1], data_300_m[1], data_150_p[1], data_150_m[1]],
                        [
                            run_Simulation_300_p, run_Simulation_300_m,
                            run_Simulation_150_p, run_Simulation_150_m
                        ]]

        fitResult = run_fit_differential_evolution(*dataSimTuple,
                                                   startParams)

        print("Fit Result:")
        print(fitResult)

        q_300_p, r_300_p = qr(run_Simulation_300_p(qzs, fitResult))
        q_300_m, r_300_m = qr(run_Simulation_300_m(qzs, fitResult))

        q_150_p, r_150_p = qr(run_Simulation_150_p(qzs, fitResult))
        q_150_m, r_150_m = qr(run_Simulation_150_m(qzs, fitResult))

        plot([q_300_p, q_300_m, q_150_p, q_150_m],
             [r_300_p, r_300_m, r_150_p, r_150_m],
             [data_300_p, data_300_m, data_150_p, data_150_m],
             [1, 1, 10, 10],
             ["300K $+$", "300K $-$", "150K $+$", "150K $-$"],
             f"Honeycomb_Fit_reflectivity_fit.pdf")

        plot_sld_profile(fitResult, f"Honeycomb_Fit_sld_profile_fit.pdf")



#!/usr/bin/env python3
"""
Fitting experimental data: spherical nanoparticles with size distribution
in 3 layers system (experiment at GALAXI).
"""
import bornagain as ba
from bornagain import nm
import ba_fitmonitor
from sample_builder import SampleBuilder
from matplotlib import pyplot as plt

wavelength = 1.34*ba.angstrom
alpha_i = 0.463*ba.deg

# detector setup as given from instrument responsible
pilatus_npx, pilatus_npy = 981, 1043
pilatus_pixel_size = 0.172  # in mm
detector_distance = 1730.0  # in mm
beam_xpos, beam_ypos = 597.1, 323.4  # in pixels


def create_detector():
    """
    Returns a model of the GALAXY detector
    """
    u0 = beam_xpos*pilatus_pixel_size  # in mm
    v0 = beam_ypos*pilatus_pixel_size  # in mm
    detector = ba.RectangularDetector(pilatus_npx,
                                      pilatus_npx*pilatus_pixel_size,
                                      pilatus_npy,
                                      pilatus_npy*pilatus_pixel_size)
    detector.setPerpendicularToDirectBeam(detector_distance, u0, v0)
    return detector


def create_simulation(params):
    """
    Creates and returns GISAS simulation with beam and detector defined
    """
    simulation = ba.GISASSimulation()
    simulation.setDetector(create_detector())
    simulation.setBeamParameters(wavelength, alpha_i, 0)
    simulation.beam().setIntensity(1.2e7)
    simulation.setRegionOfInterest(85, 70, 120, 92.)
    simulation.addMask(ba.Rectangle(101.9, 82.1, 103.7, 85.2),
                       True)  # beamstop

    sample_builder = SampleBuilder()
    sample = sample_builder.create_sample(params)
    simulation.setSample(sample)

    return simulation


def load_real_data(filename="galaxi_data.tif.gz"):
    """
    Loads experimental data and returns numpy array.
    """
    return ba.IntensityDataIOFactory.readIntensityData(filename).array()


def run_fitting():
    real_data = load_real_data()

    fit_objective = ba.FitObjective()
    fit_objective.addSimulationAndData(create_simulation, real_data, 1)
    fit_objective.initPrint(10)
    observer = ba_fitmonitor.PlotterGISAS()
    fit_objective.initPlot(10, observer)

    params = ba.Parameters()
    params.add("radius", 5.*nm, min=4, max=6, step=0.1*nm)
    params.add("sigma", 0.55, min=0.2, max=0.8, step=0.01)
    params.add("distance", 27.*nm, min=20, max=70)

    minimizer = ba.Minimizer()
    result = minimizer.minimize(fit_objective.evaluate, params)
    fit_objective.finalize(result)


if __name__ == '__main__':
    run_fitting()
    plt.show()



#!/usr/bin/env python3
"""
Reads a BornAgain simulation script, and converts into normal form.

Does not touch function get_sample, for which there is the dedicated
script normalize-getsample.py.

Export to normal form is done by BornAgain's ExportToPython function.
"""

import argparse, os, re
from yapf.yapflib.yapf_api import FormatCode
import bornagain as ba


def restitute_sample(ti, tc):
    """
    Returns script tc, except for the get_sample function which is taken from script ti.
    """
    pat = re.compile(
        r'((\ndef get_sample\(.+?:\n)(\s*""".+?"""\n)?\n*(((\s{4}.*?)?\n)+?\n)(\w|$))',
        flags=re.S)

    mi = re.search(pat, ti)
    if not mi:
        raise Exception("Input code has no function get_sample")
    header = mi.group(2)

    mn = re.search(pat, tc)
    if not mn:
        raise Exception("Normalized code has no function get_sample")
    if mn.group(2) != header:
        print(f'WARNING: Signature of function get_sample has changed from "{header}" to '
              f'"{mn.group(2)}"')

    t = re.sub(pat, mi.group(1), tc)

    return t


def retrieve_simulation(ti, fname):
    c = compile(ti, fname, 'exec')
    ns = {}
    if verbose:
        print(f'.. retrieve_simulation: compiled')
    exec(c, ns)
    if verbose:
        print(f'.. retrieve_simulation: executed')
    globals().update(ns)
    if verbose:
        print(f'.. retrieve_simulation: globals updated')
    sam = get_sample()
    if verbose:
        print(f'.. got sample')
    try:
        sim = get_simulation(sam)
        if verbose:
            print(f'.. got simulation(sam)')
    except:
        if verbose:
            print(f'.. retry get_simulation w/o argument')
        sim = get_simulation()
        if verbose:
            print(f'.. got get_simulation()')
    return sim


def cycle_text(ti, fname):
    """
    Returns normalized version of script ti as obtained from BornAgain's export-to-Python function.
    """
    s = retrieve_simulation(ti, fname)
    return ba.simulationPlotCode(s)


def normalize_text(ti, fname):
    filecomment = ""
    m = re.match(r'""".*?"""\n', ti, flags=re.S)
    if m:
        filecomment = m.group(0)

    tc = filecomment + cycle_text(ti, fname)

    if verbose:
        print(
            f'.. cycled through BA, {len(ti.split())} -> {len(tc.split())} lines'
        )
    tf = restitute_sample(ti, tc)

    if verbose:
        print(f'.. normalized, {len(ti.split())} -> {len(tf.split())} lines')
    # YAPF formatting
    tf = FormatCode(tf,
                    style_config='{based_on_style: pep8, column_limit: 80, '
                    'no_spaces_around_selected_binary_operators: "*,/"}')[0]
    return tf


def normalize_file(fname, inplace):
    try:
        print(f'FILE {fname}:')

        # read file
        with open(fname, 'r') as f:
            ti = f.read()
            if verbose:
                print(f'.. read {len(ti.split())} lines')

        m = re.search(r"""\nif __name__ == '__main__':(\n\s+.+?){0,7}
    ba_plot.run_and_plot\(.*?\)""", ti)
        if not m:
            print('=> main is not run_and_plot => SKIPPED')
            return 3

        # normalize
        tf = normalize_text(ti, fname)
        if verbose:
            print(f'.. obtained normalized text')
        if tf == ti:
            print(f'=> OK (nothing to change)')
            return 0
        with open('out1.py', 'w') as f:
            f.write(tf)
        if verbose:
            print(f'.. saved to out1.py')

        # check invariance under second normalization
        t2 = normalize_text(tf, fname)
        if verbose:
            print(f'.. re-normalized')
        if t2 != tf:
            with open('out2.py', 'w') as f:
                f.write(t2)
            exit(
                f'=> BUG - changed under second normalization, see out2.py vs out1.py'
            )

        # check invariance of simulation result:
        si = retrieve_simulation(ti, fname + ".old")
        sf = retrieve_simulation(tf, fname + ".new")
        if verbose:
            print(f'.. running old simulation ..')
        si.runSimulation()
        ri = si.result()
        if verbose:
            print(f'.. running new simulation ..')
        sf.runSimulation()
        rf = sf.result()
        if verbose:
            print(f'.. done with simulations')
        diff = ba.relativeDataDifference(ba.importArrayToOutputData(ri.array()),
                                         ba.importArrayToOutputData(rf.array()))
        if verbose:
            print(f'.. relative difference {diff}')
        if diff > 1e-10:
            exit(f'=> BUG - simulation result changed')

        # output
        if inplace:
            os.rename('out1.py', fname)
            print(f'=> NORMALIZED')
        else:
            print(tf)

        return 1

    except Exception as e:
        print(f'=> FAILED - {e}')
        return 2


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("input_files", nargs='+', type=str)
    parser.add_argument("-i", "--in-place", action="store_true")
    parser.add_argument("-v", "--verbose", action="store_true")
    args = parser.parse_args()
    verbose = args.verbose
    files = args.input_files

    count = [0, 0, 0, 0]
    for f in files:
        ret = normalize_file(f, args.in_place)
        count[ret] += 1

    if (len(files) <= 1):
        exit(0)

    out = []
    if count[0] > 0:
        out.append(f'{count[0]} unchanged')
    if count[1] > 0:
        out.append(f'{count[1]} normalized')
    if count[2] > 0:
        out.append(f'{count[2]} failed')
    if count[2] > 0:
        out.append(f'{count[3]} skipped')
    print(f'TOTAL of {len(args.input_files)} files: {", ".join(out)}')



# Functional test: functional test: mesocrystal simulation
#

from __future__ import print_function
import ctypes, math, numpy, os, sys, time

import utils
from bornagain import *


# ----------------------------------------------------------------------------
# Sample builder to build mixture of cylinders and prisms on top of substrate
# ----------------------------------------------------------------------------
class MySampleBuilder(ISampleBuilder):
    def __init__(self):
        ISampleBuilder.__init__(self)
        self.sample = None
        # parameters describing the sample
        self.lattice_length_a = ctypes.c_double(6.2091e+00*nm)
        self.lattice_length_c = ctypes.c_double(6.5677e+00*nm)
        self.nanoparticle_radius = ctypes.c_double(4.6976e+00*nm)
        self.sigma_nanoparticle_radius = ctypes.c_double(3.6720e-01*nm)
        self.meso_height = ctypes.c_double(1.1221e+02*nm)
        self.meso_radius = ctypes.c_double(9.4567e+02*nm)
        self.sigma_lattice_length_a = ctypes.c_double(1.1601e+00*nm)
        self.surface_filling_ratio = ctypes.c_double(1.7286e-01)
        self.roughness = ctypes.c_double(2.8746e+01*nm)
        # register parameters
        self.registerParameter(
            "lattice_length_a", ctypes.addressof(
                self.lattice_length_a)).setUnit("nm").setNonnegative()
        self.registerParameter(
            "lattice_length_c", ctypes.addressof(
                self.lattice_length_c)).setUnit("nm").setNonnegative()
        self.registerParameter(
            "nanoparticle_radius",
            ctypes.addressof(
                self.nanoparticle_radius)).setUnit("nm").setNonnegative()
        self.registerParameter(
            "sigma_nanoparticle_radius",
            ctypes.addressof(self.sigma_nanoparticle_radius)).setUnit(
                "nm").setNonnegative()
        self.registerParameter(
            "meso_height", ctypes.addressof(
                self.meso_height)).setUnit("nm").setNonnegative()
        self.registerParameter(
            "meso_radius", ctypes.addressof(
                self.meso_radius)).setUnit("nm").setNonnegative()
        self.registerParameter(
            "sigma_lattice_length_a",
            ctypes.addressof(self.sigma_lattice_length_a)).setUnit(
                "nm").setNonnegative()
        self.registerParameter(
            "surface_filling_ratio",
            ctypes.addressof(self.surface_filling_ratio)).setNonnegative()
        self.registerParameter("roughness", ctypes.addressof(
            self.roughness)).setUnit("nm").setNonnegative()

    # -------------------------------------------------------------------------
    # constructs the sample for current values of parameters
    # -------------------------------------------------------------------------
    def buildSample(self):
        surface_density = self.surface_filling_ratio.value/numpy.pi/self.meso_radius.value/self.meso_radius.value
        n_particle = complex(1.0 - 2.84e-5, 4.7e-7)
        avg_n_squared_meso = complex(0.7886*n_particle*n_particle + 0.2114)
        n_avg = complex(
            numpy.sqrt(self.surface_filling_ratio.value*
                       avg_n_squared_meso + 1.0 -
                       self.surface_filling_ratio.value))
        n_particle_adapted = complex(
            numpy.sqrt(n_avg*n_avg + n_particle*n_particle - 1))
        ff = FormFactorCylinder(self.meso_radius.value,
                                self.meso_height.value)

        # Create multilayer
        p_multi_layer = MultiLayer()
        n_air = complex(1, 0)
        n_substrate = complex(1.0 - 7.57e-6, 1.73e-7)

        p_vacuum_material = HomogeneousMaterial("Vacuum", n_air)
        p_average_layer_material = HomogeneousMaterial(
            "Averagelayer", n_avg)
        p_substrate_material = HomogeneousMaterial("Substrate",
                                                   n_substrate)
        vacuum_layer = Layer(p_vacuum_material)
        avg_layer = Layer(p_average_layer_material, self.meso_height.value)
        substrate_layer = Layer(p_substrate_material)
        p_interference_function = InterferenceFunctionNone()
        particle_layout = ParticleLayout()

        n_max_phi_rotation_steps = 2
        n_alpha_rotation_steps = 1
        alpha_step = 5*deg/n_alpha_rotation_steps
        alpha_start = -(n_alpha_rotation_steps/2)*alpha_step

        phi_step = 2*numpy.pi/3.0/n_max_phi_rotation_steps
        phi_start = 0.0
        for i in range(0, n_max_phi_rotation_steps):
            for j in range(0, n_alpha_rotation_steps):

                total_transform = RotationZ(phi_start + i*phi_step)
                meso = self.createMesoCrystal(self.lattice_length_a.value,
                                              self.lattice_length_c.value,
                                              n_particle_adapted, ff)
                meso.setPosition(0, 0, -self.meso_height.value)
                particle_layout.addParticle(meso, 1, kvector_t(0, 0, 0),
                                            total_transform)

        particle_layout.setTotalParticleSurfaceDensity(surface_density)
        particle_layout.setInterferenceFunction(p_interference_function)

        avg_layer.addLayout(particle_layout)

        roughness = LayerRoughness(self.roughness.value, 0.3, 500*nm)

        p_multi_layer.addLayer(vacuum_layer)
        p_multi_layer.addLayer(avg_layer)
        p_multi_layer.addLayerWithTopRoughness(substrate_layer, roughness)

        return p_multi_layer

    # -------------------------------------------------------------------------
    # building meso crystal
    # -------------------------------------------------------------------------
    def createMesoCrystal(self, stacking_radius_a, stacking_radius_c,
                          n_particle, p_meso_form_factor):

        mParticle = HomogeneousMaterial("Particle", n_particle)

        p_lat = self.createLattice(stacking_radius_a, stacking_radius_c)
        bas_a = p_lat.getBasisVectorA()
        bas_b = p_lat.getBasisVectorB()
        bas_c = p_lat.getBasisVectorC()
        ff = FormFactorSphereGaussianRadius(
            self.nanoparticle_radius.value,
            self.sigma_nanoparticle_radius.value)
        particle = Particle(mParticle, ff)
        position_0 = kvector_t(0, 0, 0)
        position_1 = 1.0/3*(2*bas_a + bas_b + bas_c)
        position_2 = 1.0/3*(bas_a + 2*bas_b + 2*bas_c)
        positions = [position_0, position_1, position_2]
        basis = ParticleComposition()
        basis.addParticles(particle, positions)

        position_variance = self.sigma_lattice_length_a.value*self.sigma_lattice_length_a.value/3.0
        npc = Crystal(basis, p_lat, position_variance)
        meso = MesoCrystal(npc, p_meso_form_factor)
        return meso

    # -------------------------------------------------------------------------
    # create lattice
    # -------------------------------------------------------------------------
    def createLattice(self, stacking_radius_a, stacking_radius_c):
        result = HexagonalLattice(stacking_radius_a*2,
                                  stacking_radius_c*2*2.3)
        result.setSelectionRule(SimpleSelectionRule(-1, 1, 1, 3))
        return result


# -----------------------------------------------------------------------------
# run mesocrystal simulation
# -----------------------------------------------------------------------------
def runTest():
    # setting simulation
    sample_builder = MySampleBuilder()
    simulation = createSimulation()
    simulation.setSampleBuilder(sample_builder)

    reference = utils.get_reference_data("mesocrystal01_reference.int.gz")

    simulation.runSimulation()
    result = simulation.result().histogram2d()

    diff = getRelativeDifference(result, reference)

    status = "OK"
    if diff > 1e-10 or numpy.isnan(diff):
        status = "FAILED"
    return "MesoCrystal1", "Mesocrystal simulation", diff, status


# create simulation
def createSimulation():
    simulation = GISASSimulation()
    simulation.setBeamParameters(1.77*angstrom, 0.4*deg, 0)
    simulation.beam().setIntensity(5.0090e+12)
    simulation.setDetectorParameters(50, 0.2*deg, 2.5*deg, 50, 0, 2.5*deg)
    return simulation


if __name__ == '__main__':
    name, description, diff, status = runTest()
    print(name, description, diff, status)
    if "FAILED" in status:
        exit(1)



"""
Functional test to check slicing mechanism for spherical particles
when they are crossing an interface.
"""

from __future__ import print_function
import os, sys, unittest

import utils
import bornagain as ba
from bornagain import deg, kvector_t

mSubstrate = ba.HomogeneousMaterial("Substrate", 3.212e-6, 3.244e-8)
mAmbience = ba.HomogeneousMaterial("Vacuum", 0, 0)


class SlicedSpheresTest(unittest.TestCase):
    def get_sample(self, particle_to_air=None, particle_to_substrate=None):
        """
        Helper function returning a multilayer (air, substrate) using particles provided
        by the user.
        """

        vacuum_layer = ba.Layer(mAmbience)
        if particle_to_air:
            layout = ba.ParticleLayout()
            layout.addParticle(particle_to_air)
            vacuum_layer.addLayout(layout)

        substrate = ba.Layer(mSubstrate)
        if particle_to_substrate:
            layout = ba.ParticleLayout()
            layout.addParticle(particle_to_substrate)
            substrate.addLayout(layout)

        multi_layer = ba.MultiLayer()
        multi_layer.addLayer(vacuum_layer)
        multi_layer.addLayer(substrate)
        return multi_layer

    def get_result(self, particle_to_air=None, particle_to_substrate=None):
        sample = self.get_sample(particle_to_air, particle_to_substrate)
        simulation = utils.get_simulation_MiniGISAS(sample)
        simulation.runSimulation()
        return simulation.result()

    def testSphericalCupOnTopOfSubstrate(self):
        """
        Compares two simulation intended to  provide identical results.
        Simulation #1: truncated sphere on top of substrate.
        Simulation #2: spherical particle crossing the interface.
        Both particles are made of same material as substrate. From scattering point of view,
        both cases should look like a truncated sphere on top of substrate.
        """

        sphere_radius = 10.0
        sphere_shift = 4.0  # shift beneath interface in absolute units

        # truncated sphere on top of substrate with height 16nm
        truncatedSphere = ba.Particle(
            mSubstrate,
            ba.FormFactorTruncatedSphere(sphere_radius,
                                         sphere_radius*2 - sphere_shift,
                                         0))
        reference = self.get_result(truncatedSphere)

        # sphere crossing interface to look like truncated sphere above
        sphere = ba.Particle(mSubstrate,
                             ba.FormFactorFullSphere(sphere_radius))
        sphere.setPosition(0, 0, -sphere_shift)
        data = self.get_result(sphere)

        diff = ba.RelativeDifference(data, reference)
        print(diff)
        self.assertLess(diff, 1e-10)

    def testSphericalLacuneInSubstrate(self):
        """
        Similar to previous. Truncated sphere and sphere are made of vacuum material.
        From scattering point of view, both cases should look like an vacuum lacune in substrate.
        """

        sphere_radius = 10.0
        sphere_shift = 4.0  # shift beneath interface in absolute units

        # Sphere truncated from top. Intended to go below interface.
        truncatedSphere = ba.Particle(
            mAmbience,
            ba.FormFactorTruncatedSphere(sphere_radius, sphere_radius*2,
                                         sphere_radius*2 - sphere_shift))
        truncatedSphere.setPosition(0, 0, -sphere_shift)
        reference = self.get_result(truncatedSphere)

        # sphere crossing interface to look like truncated sphere above
        sphere = ba.Particle(mAmbience,
                             ba.FormFactorFullSphere(sphere_radius))
        sphere.setPosition(0, 0, -sphere_shift)
        data = self.get_result(sphere)

        diff = ba.RelativeDifference(data, reference)
        print(diff)
        self.assertLess(diff, 1e-10)

    def testSpheresCrossingInterface(self):
        """
        Compares two simulation intended to  provide identical results.
        Simulation #1: full sphere inserted in vacuum layer to cross interface
        Simulation #2: full sphere inserted in substrate layer to cross interface
        Both spheres are made of same material.
        """
        mParticle = ba.HomogeneousMaterial("Ag", 1.245e-5, 5.419e-7)

        sphere_radius = 10.0
        sphere_shift = 4.0  # shift beneath interface in absolute units

        # Sphere intended for vacuum layer and crossing interface
        sphere1 = ba.Particle(mParticle,
                              ba.FormFactorFullSphere(sphere_radius))
        sphere1.setPosition(0, 0, -sphere_shift)
        reference = self.get_result(particle_to_air=sphere1)

        # Sphere intended for substrate layer and crossing interface
        sphere2 = ba.Particle(mParticle,
                              ba.FormFactorFullSphere(sphere_radius))
        sphere2.setPosition(0, 0, -sphere_shift)
        data = self.get_result(particle_to_substrate=sphere2)

        diff = ba.RelativeDifference(data, reference)
        print(diff)
        self.assertLess(diff, 1e-10)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
Cylindrical particle made from two materials.
Particle crosses air/substrate interface.
"""
import bornagain as ba
from bornagain import deg, nm, kvector_t


def get_sample():
    """
    Returns a multi layer with substrate/air layers.
    Vacuum layer contains cylindrical particles made of two materials.
    Particle shifted down to cross interface.
    """

    # Define materials
    material_Ag = ba.HomogeneousMaterial("Ag", 1.245e-05, 5.419e-07)
    material_Substrate = ba.HomogeneousMaterial("Substrate", 3.212e-06,
                                                3.244e-08)
    material_Teflon = ba.HomogeneousMaterial("Teflon", 2.9e-06, 6.019e-09)
    material_Vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)

    # Define form factors
    ff_1 = ba.FormFactorCylinder(10*nm, 4*nm)
    ff_2 = ba.FormFactorCylinder(10*nm, 10*nm)

    # Define particles
    subparticle_1 = ba.Particle(material_Ag, ff_1)
    subparticle_1.setPosition(kvector_t(0, 0, 10*nm))
    subparticle_2 = ba.Particle(material_Teflon, ff_2)

    # Define composition of particles at specific positions
    particle = ba.ParticleComposition()
    particle.addParticle(subparticle_1)
    particle.addParticle(subparticle_2)
    particle.setPosition(kvector_t(0, 0, -10*nm))

    # Define particle layouts
    layout = ba.ParticleLayout()
    layout.addParticle(particle)
    layout.setTotalParticleSurfaceDensity(1)

    # Define layers
    layer_1 = ba.Layer(material_Vacuum)
    layer_1.addLayout(layout)
    layer_2 = ba.Layer(material_Substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1e9, 0.1*nm, ba.Direction(0.2*deg, 0))
    detector = ba.SphericalDetector(100, 2*deg, 0, 1*deg)
    simulation = ba.GISASSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)
